# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the scraper functions for CASTEP-related
inputs and outputs.

"""


from collections import defaultdict
import os
import glob
import gzip
import warnings
import pwd
import numpy as np
from matador.utils.cell_utils import abc2cart, calc_mp_spacing, cart2volume, wrap_frac_coords, cart2abc, cart2frac
from matador.utils.chem_utils import get_stoich, INVERSE_CM_TO_EV
from matador.scrapers.utils import DFTError, ComputationError, scraper_function, f90_float_parse


@scraper_function
def res2dict(seed, db=True, **kwargs):
    """ Extract available information from .res file; preferably
    used in conjunction with cell or param file.

    Parameters:
        seed (str or list): filename or list of filenames of res file(s)
            (with or without extension).

    Keyword arguments:
        db (bool): whether to fail if unable to scrape energies.

    Returns:
        (tuple): containing either dict/str containing data or error, and a bool stating
            if the scrape was successful.

    """
    res = dict()
    # read .res file into array
    if seed.endswith('.res'):
        seed = seed.replace('.res', '')
    with open(seed + '.res', 'r') as f:
        flines = f.readlines()
    # add .res to source
    res['source'] = []
    res['source'].append(seed + '.res')
    # grab file owner username
    try:
        res['user'] = pwd.getpwuid(os.stat(seed + '.res').st_uid).pw_name
    except Exception:
        pass

    get_seed_metadata(res, seed)
    # alias special lines in res file
    titl = ''
    cell = ''
    remark = ''
    for line in flines:
        if 'TITL' in line and db:
            # if not db, then don't read title
            titl = line.split()
            if len(titl) != 12:
                raise RuntimeError('missing some TITL info')
        elif 'CELL' in line:
            cell = line.split()
        elif 'REM' in line:
            remark = line.split()
    if cell == '':
        raise RuntimeError('missing CELL info')
    if titl == '' and db:
        raise RuntimeError('missing TITL')
    if db:
        res['pressure'] = f90_float_parse(titl[2])
        res['cell_volume'] = f90_float_parse(titl[3])
        res['enthalpy'] = f90_float_parse(titl[4])
        res['num_atoms'] = int(titl[7])
        res['space_group'] = titl[8].strip('()')
        res['enthalpy_per_atom'] = res['enthalpy'] / res['num_atoms']
    res['lattice_abc'] = [list(map(f90_float_parse, cell[2:5])), list(map(f90_float_parse, cell[5:8]))]
    # calculate lattice_cart from abc
    res['lattice_cart'] = abc2cart(res['lattice_abc'])
    if 'cell_volume' not in res:
        res['cell_volume'] = cart2volume(res['lattice_cart'])
    res['atom_types'] = []
    res['positions_frac'] = []
    res['site_occupancy'] = []
    for line_no, line in enumerate(flines):
        if 'SFAC' in line:
            i = 1
            while 'END' not in flines[line_no + i] and line_no + i < len(flines):
                # check if we don't have some other SHELX keyword in the way, e.g. "UNIT"
                if(len(flines[line_no + i]) >= 4 and
                   all([char.isupper() for char in flines[line_no+i][0:4]])):
                    i += 1
                    continue
                cursor = flines[line_no + i].split()
                res['atom_types'].append(cursor[0])
                res['positions_frac'].append(list(map(f90_float_parse, cursor[2:5])))
                try:
                    res['site_occupancy'].append(f90_float_parse(cursor[5]))
                except IndexError:
                    res['site_occupancy'].append(1.0)
                i += 1
    res['positions_frac'] = wrap_frac_coords(res['positions_frac'])
    res['num_atoms'] = len(res['atom_types'])
    # deal with implicit encapsulation
    if remark:
        if 'NTPROPS' in remark:
            res['cnt_chiral'] = [0, 0]
            res['encapsulated'] = True
            for ind, entry in enumerate(remark):
                if 'chiralN' in entry:
                    res['cnt_chiral'][0] = int(remark[ind + 1].replace(',', ''))
                if 'chiralM' in entry:
                    res['cnt_chiral'][1] = int(remark[ind + 1].replace(',', ''))
                if entry == '\'r\':':
                    res['cnt_radius'] = f90_float_parse(remark[ind + 1].replace(',', ''))
                if entry == '\'z\':':
                    temp_length = remark[ind + 1].replace(',', '')
                    temp_length = temp_length.replace('\n', '')
                    temp_length = temp_length.replace('}', '')
                    res['cnt_length'] = f90_float_parse(temp_length)

    res['stoichiometry'] = get_stoich(res['atom_types'])
    res['num_fu'] = len(res['atom_types']) / sum([elem[1] for elem in res['stoichiometry']])

    return res, True


@scraper_function
def cell2dict(seed, db=False, lattice=True, positions=True, **kwargs):
    """ Extract available information from .cell file; probably
    to be merged with another dict from a .param or .res file.

    Parameters:
        seed (str/list): filename or list of filenames of cell file(s)
            to scrape, with or without extension.

    Keyword arguments:
        db (bool): scrape database quality file
        lattice (bool): scrape lattice vectors
        positions (bool): scrape positions

    Returns:
        (tuple): containing either dict/str containing data or error, and a bool stating
            if the scrape was successful.

    """
    cell = dict()
    if seed.endswith('.cell'):
        seed = seed.replace('.cell', '')
    with open(seed + '.cell', 'r') as f:
        flines = f.readlines()
    # add cell file to source
    cell['source'] = []
    cell['source'].append(seed + '.cell')
    for line_no, line in enumerate(flines):
        if line.startswith(('#', '!')):
            continue
        if '#' or '!' in line:
            line = line.split('#')[0].split('!')[0]
        if '%block lattice_cart' in line.lower() and lattice:
            cell['lattice_cart'] = []
            i = 1
            while 'endblock' not in flines[line_no + i].lower():
                if not flines[line_no + i].strip()[0].isalpha():
                    cell['lattice_cart'].append(list(map(f90_float_parse, flines[line_no + i].split())))
                    if not len(cell['lattice_cart'][-1]) == 3:
                        raise RuntimeError('Lattice vector does not have enough elements!')
                i += 1
            if not len(cell['lattice_cart']) == 3:
                raise RuntimeError('Wrong number of lattice vectors!')
        elif '%block lattice_abc' in line.lower() and lattice:
            cell['lattice_abc'] = []
            i = 1
            while 'endblock' not in flines[line_no + i].lower():
                if not flines[line_no + i].strip()[0].isalpha():
                    cell['lattice_abc'].append(list(map(f90_float_parse, flines[line_no + i].split())))
                    if not len(cell['lattice_abc'][-1]) == 3:
                        raise RuntimeError('Lattice vector does not have enough elements!')
                i += 1
            if not len(cell['lattice_abc']) == 2:
                raise RuntimeError('Wrong specification of lattice_abc')

        elif '%block species_pot' in line.lower():
            cell['species_pot'] = dict()
            i = 1
            while 'endblock' not in flines[line_no + i].lower():
                if db:
                    species = flines[line_no+i].split()[0]
                    pspot_string = flines[line_no+i].split()[1].split('/')[-1]
                    cell['species_pot'][species] = pspot_string.replace('()', '').replace('[]', '')
                else:
                    pspot_libs = ['C7', 'C8', 'C9', 'C17', 'C18', 'MS', 'HARD',
                                  'QC5', 'NCP', 'NCP18', 'NCP17', 'NCP9']
                    if flines[line_no + i].upper().split()[0] in pspot_libs:
                        cell['species_pot']['library'] = flines[line_no + i].upper().split()[0]
                    else:
                        cell['species_pot'][flines[line_no + i].split()[0]] = flines[line_no + i].split()[1]
                i += 1
        elif '%block cell_constraints' in line.lower():
            cell['cell_constraints'] = []
            for j in range(2):
                cell['cell_constraints'].append(list(map(int, flines[line_no + j + 1].split())))
            if (any(len(cell['cell_constraints'][i]) != 3 for i in range(2))
                    or len(cell['cell_constraints']) != 2):
                raise RuntimeError('Invalid cell constraints block.')
        elif '%block hubbard_u' in line.lower():
            cell['hubbard_u'] = defaultdict(list)
            i = 0
            while 'endblock' not in flines[line_no + i].lower():
                line = flines[line_no + i]
                if line == 'eV' or len(line.split()) < 3:
                    i += 1
                    continue
                else:
                    atom = line.split()[0]
                    orbital = line.split()[1].replace(':', '')
                    shift = f90_float_parse(line.split()[-1])
                    atom = line.split()[0]
                    cell['hubbard_u'][atom] = dict()
                    cell['hubbard_u'][atom][orbital] = shift
                    i += 1
        elif '%block external_pressure' in line.lower():
            cell['external_pressure'] = np.zeros((3, 3))
            i = 1
            j = 0
            while 'endblock' not in flines[line_no+i].lower():
                if not flines[line_no+i].strip()[0].isalpha():
                    flines[line_no+i] = flines[line_no+i].replace(',', '')
                    vals = list(map(f90_float_parse, flines[line_no+i].split()))
                    if len(vals) != (3-j):
                        raise RuntimeError('External pressure should be specified as upper triangular matrix.')
                    cell['external_pressure'][j] = np.asarray(j*[0.0] + vals).reshape(3)
                    j += 1
                i += 1
            cell['external_pressure'] = cell['external_pressure'].tolist()
        elif '%block ionic_constraints' in line.lower():
            cell['ionic_constraints'] = []
            i = 1
            while 'endblock' not in flines[line_no+i].lower():
                cell['ionic_constraints'].append(flines[line_no+i].strip())
                i += 1
        # parse kpoints
        elif 'kpoints_mp_spacing' in line.lower() or 'kpoint_mp_spacing' in line.lower():
            if 'spectral_kpoints_mp_spacing' in line.lower() or 'spectral_kpoint_mp_spacing' in line.lower():
                cell['spectral_kpoints_mp_spacing'] = f90_float_parse(line.split()[-1])
            elif 'supercell_kpoints_mp_spacing' in line.lower() or 'supercell_kpoint_mp_spacing' in line.lower():
                cell['supercell_kpoints_mp_spacing'] = f90_float_parse(line.split()[-1])
            elif 'phonon_kpoints_mp_spacing' in line.lower() or 'phonon_kpoint_mp_spacing' in line.lower():
                cell['phonon_kpoint_mp_spacing'] = f90_float_parse(line.split()[-1])
            elif 'phonon_fine_kpoints_mp_spacing' in line.lower() or 'phonon_fine_kpoint_mp_spacing' in line.lower():
                cell['phonon_fine_kpoint_mp_spacing'] = f90_float_parse(line.split()[-1])
            else:
                cell['kpoints_mp_spacing'] = f90_float_parse(line.split()[-1])
        elif 'kpoints_mp_grid' in line.lower() or 'kpoint_mp_grid' in line.lower():
            if 'spectral_kpoints_mp_grid' in line.lower() or 'spectral_kpoint_mp_grid' in line.lower():
                cell['spectral_kpoints_mp_grid'] = list(map(int, line.split()[-3:]))
            # these two keywords do not have a corresponding "kpoints" alias (instead of kpoint)
            # so must remain unpluralised (see below for other phonon keyword exceptions)
            elif 'phonon_kpoints_mp_grid' in line.lower() or 'phonon_kpoint_mp_grid' in line.lower():
                cell['phonon_kpoint_mp_grid'] = list(map(int, line.split()[-3:]))
            elif 'phonon_fine_kpoints_mp_grid' in line.lower() or 'phonon_fine_kpoint_mp_grid' in line.lower():
                cell['phonon_fine_kpoint_mp_grid'] = list(map(int, line.split()[-3:]))
            else:
                cell['kpoints_mp_grid'] = list(map(int, line.split()[-3:]))
        elif 'kpoints_mp_offset' in line.lower() or 'kpoint_mp_offset' in line.lower():
            if 'spectral_kpoints_mp_offset' in line.lower() or 'spectral_kpoint_mp_offset' in line.lower():
                cell['spectral_kpoints_mp_offset'] = list(map(f90_float_parse, line.split()[-3:]))
            elif 'phonon_kpoints_mp_offset' in line.lower() or 'phonon_kpoint_mp_offset' in line.lower():
                # this is a special case where phonon_kpointS_mp_offset doesn't exist
                cell['phonon_kpoint_mp_offset'] = list(map(f90_float_parse, line.split()[-3:]))
            elif 'phonon_fine_kpoints_mp_offset' in line.lower() or 'phonon_fine_kpoint_mp_offset' in line.lower():
                cell['phonon_fine_kpoint_mp_offset'] = list(map(f90_float_parse, line.split()[-3:]))
            else:
                cell['kpoints_mp_offset'] = list(map(f90_float_parse, line.split()[-3:]))
        elif '%block' in line.lower() and ('_kpoints_path' in line.lower() or '_kpoint_path' in line.lower()):
            if 'spectral' in line.lower():
                key = 'spectral_kpoints_path'
            elif 'phonon_fine' in line.lower():
                key = 'phonon_fine_kpoint_path'
            else:
                raise RuntimeError('Found unknown kpoint path block.')
            i = 1
            labels = []
            found_labels = False
            comment_delims = ['!', '%', '#']
            cell[key] = []
            while '%endblock' not in flines[line_no + i].lower():
                for delim in comment_delims:
                    if delim in flines[line_no+i]:
                        labels.append(flines[line_no+i].split(delim)[-1].strip())
                        found_labels = True
                        break
                else:
                    labels.append(None)

                cell[key].append(list(map(f90_float_parse, flines[line_no + i].split()[:3])))
                i += 1
            if found_labels:
                cell[key + '_labels'] = labels
        elif '%block spectral_kpoints_list' in line.lower() or '%block spectral_kpoint_list' in line.lower():
            i = 1
            cell['spectral_kpoints_list'] = []
            while '%endblock' not in flines[line_no + i].lower():
                cell['spectral_kpoints_list'].append(list(map(f90_float_parse, flines[line_no + i].split()[:4])))
                i += 1
        elif '%block phonon_fine_kpoints_list' in line.lower() or '%block phonon_fine_kpoint_list' in line.lower():
            i = 1
            # this is a special case where phonon_fine_kpointS_list doesn't exist
            cell['phonon_fine_kpoint_list'] = []
            while '%endblock' not in flines[line_no + i].lower():
                cell['phonon_fine_kpoint_list'].append(list(map(f90_float_parse, flines[line_no + i].split()[:4])))
                i += 1
        elif '%block phonon_supercell_matrix' in line.lower():
            cell['phonon_supercell_matrix'] = []
            i = 1
            while 'endblock' not in flines[line_no + i].lower():
                cell['phonon_supercell_matrix'].append(list(map(int, flines[line_no + i].split())))
                if not len(cell['phonon_supercell_matrix'][-1]) == 3:
                    raise RuntimeError('Supercell matrix row does not have enough elements!')
                i += 1
            if not len(cell['phonon_supercell_matrix']) == 3:
                raise RuntimeError('Wrong supercell matrix shape!')

        elif not db:
            if '%block positions_frac' in line.lower():
                atomic_init_spins = []
                i = 1
                if positions:
                    cell['atom_types'] = []
                    cell['positions_frac'] = []
                while '%endblock positions_frac' not in flines[line_no + i].lower():
                    line = flines[line_no + i].split()
                    if positions:
                        cell['atom_types'].append(line[0])
                        cell['positions_frac'].append(list(map(f90_float_parse, line[1:4])))
                    if 'spin=' in flines[line_no + i].lower():
                        split_line = flines[line_no + i].split()
                        atomic_init_spins.append(float(split_line[-1].lower().replace('spin=', '')))
                    else:
                        atomic_init_spins.append(None)
                    i += 1
                if any(atomic_init_spins):
                    cell['atomic_init_spins'] = atomic_init_spins
                    if len(cell['atomic_init_spins']) != len(cell['positions_frac']):
                        raise RuntimeError('Atomic init spins do not match positions')
                if positions:
                    cell['positions_frac'] = wrap_frac_coords(cell['positions_frac'])
                    cell['num_atoms'] = len(cell['atom_types'])
            elif '%block positions_abs' in line.lower():
                atomic_init_spins = []
                i = 1
                # avoid units
                if len(flines[line_no+i].split()) < 3:
                    i += 1
                if positions:
                    cell['atom_types'] = []
                    cell['positions_abs'] = []
                while '%endblock positions_abs' not in flines[line_no + i].lower():
                    line = flines[line_no + i].split()
                    if positions:
                        cell['atom_types'].append(line[0])
                        cell['positions_abs'].append(list(map(f90_float_parse, line[1:4])))
                    if 'spin=' in flines[line_no + i].lower():
                        split_line = flines[line_no + i].split()
                        atomic_init_spins.append(float(split_line[-1].lower().replace('spin=', '')))
                    else:
                        atomic_init_spins.append(None)
                    i += 1
                if any(atomic_init_spins):
                    cell['atomic_init_spins'] = atomic_init_spins
                    if len(cell['atomic_init_spins']) != len(cell['positions_frac']):
                        raise RuntimeError('Atomic init spins do not match positions')

            elif 'fix_com' in line.lower():
                cell['fix_com'] = bool(line.split()[-1])
            elif 'fix_all_ions' in line.lower():
                cell['fix_all_ions'] = bool(line.split()[-1])
            elif 'fix_all_cell' in line.lower():
                cell['fix_all_cell'] = bool(line.split()[-1])
            elif 'fix_vol' in line.lower():
                cell['fix_vol'] = bool(line.split()[-1])
            elif 'symmetry_generate' in line.lower():
                cell['symmetry_generate'] = True
            elif 'symmetry_tol' in line.lower():
                cell['symmetry_tol'] = f90_float_parse(line.split()[-1])
            elif 'snap_to_symmetry' in line.lower():
                cell['snap_to_symmetry'] = True
            elif 'quantisation_axis' in line.lower():
                cell['quantisation_axis'] = list(map(int, line.split()[1:]))
            elif 'positions_noise' in line.lower():
                cell['positions_noise'] = f90_float_parse(line.split()[-1])
            elif 'cell_noise' in line.lower():
                cell['cell_noise'] = f90_float_parse(line.split()[-1])
            elif 'kpoints_path' in line.lower() or 'kpoint_path' in line.lower() and '%block' not in line.lower():
                if 'spectral_kpoints_path_spacing' in line.lower() or 'spectral_kpoint_path_spacing' in line.lower():
                    cell['spectral_kpoints_path_spacing'] = f90_float_parse(line.split()[-1])
                elif 'phonon_fine_kpoints_path_spacing' in line.lower() or 'phonon_fine_kpoint_path_spacing' in line.lower():
                    cell['phonon_fine_kpoint_path_spacing'] = f90_float_parse(line.split()[-1])
                elif 'kpoints_path_spacing' in line.lower() or 'kpoint_path_spacing' in line.lower():
                    cell['kpoints_path_spacing'] = f90_float_parse(line.split()[-1])

    if 'external_pressure' not in cell or not cell['external_pressure']:
        cell['external_pressure'] = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]

    if lattice:
        if 'lattice_cart' not in cell and 'lattice_abc' in cell:
            cell['lattice_cart'] = abc2cart(cell['lattice_abc'])
        elif 'lattice_cart' in cell and 'lattice_abc' not in cell:
            cell['lattice_abc'] = cart2abc(cell['lattice_cart'])
        if 'lattice_cart' in cell:
            cell['cell_volume'] = cart2volume(cell['lattice_cart'])

    if positions:
        if 'positions_frac' not in cell and 'positions_abs' in cell:
            cell['positions_frac'] = cart2frac(cell['lattice_cart'], cell['positions_abs'])
            cell['positions_frac'] = wrap_frac_coords(cell['positions_frac'])

    if db:
        for species in cell['species_pot']:
            if 'OTF' in cell['species_pot'][species].upper():
                pspot_seed = '/'.join(seed.split('/')[:-1]) + '/' + cell['species_pot'][species]
                if os.path.isfile(pspot_seed):
                    cell['species_pot'].update(usp2dict(pspot_seed))

    return cell, True


@scraper_function
def param2dict(seed, db=True, **kwargs):
    """ Extract available information from .param file; probably
    to be merged with other dicts from other files.

    Parameters:
        seed (str/list): param filename or list of filenames with or
            without file extension

    Keyword arguments:
        db (bool): if True, only scrape relevant info, otherwise scrape all

    Returns:
        (tuple): containing either dict/str containing data or error, and a bool stating
            if the scrape was successful.

    """
    from matador.utils.castep_params import CASTEP_PARAMS
    param = dict()
    if seed.endswith('.param'):
        seed = seed.replace('.param', '')
    with open(seed + '.param', 'r') as f:
        flines = f.readlines()
    param['source'] = []
    param['source'].append(seed + '.param')
    # exclude some useless info if importing to db
    scrub_list = ['checkpoint', 'write_bib', 'mix_history_length',
                  'fix_occupancy', 'page_wvfns', 'num_dump_cycles',
                  'backup_interval', 'geom_max_iter', 'fixed_npw',
                  'write_cell_structure', 'bs_write_eigenvalues',
                  'calculate_stress', 'opt_strategy', 'max_scf_cycles']
    splitters = [':', '=', '\t', ' ']
    unrecognised = []
    devel_lines = []
    for line_no, line in enumerate(flines):
        if '#' or '!' in line:
            line = line.split('#')[0].split('!')[0]
        line = line.lower()
        # skip blank lines and comments
        if line.startswith(('#', '!')) or not line.strip() or line_no in devel_lines:
            continue
        else:
            # if scraping to db, ignore "rubbish"
            if db:
                if [rubbish for rubbish in scrub_list if rubbish in line]:
                    continue
            # read all other parameters in
            for splitter in splitters:
                if splitter in line:
                    if not line.startswith('%'):
                        keyword = line.split(splitter)[0].strip()
                        value = line.split(splitter)[-1].strip()
                        if keyword.lower() not in CASTEP_PARAMS:
                            unrecognised.append(keyword.lower())
                        param[keyword] = value
                    # deal with edge cases
                    if '%block devel_code' in line:
                        i = 1
                        while '%endblock devel_code' not in flines[line_no+i].lower():
                            if i+line_no >= len(flines):
                                raise RuntimeError('Found unclosed %block devel_code.')
                            line = flines[line_no+i].lower()
                            devel_lines.append(line_no+i)
                            if 'devel_code' not in param:
                                param['devel_code'] = ''
                            param['devel_code'] += line
                            i += 1
                        break

                    if 'true' in value.lower():
                        param[keyword] = True
                    elif 'false' in value.lower():
                        param[keyword] = False

                    if 'spin_polarised' in line:
                        param['spin_polarized'] = param['spin_polarised']
                        if 'spin_polarised' in param:
                            del param['spin_polarised']

                    if keyword == 'geom_max_iter':
                        param['geom_max_iter'] = int(param['geom_max_iter'])

                    if 'cut_off_energy' in line and 'mix_cut_off_energy' not in line:
                        temp_cut_off = param['cut_off_energy'].split()
                        if len(temp_cut_off) > 1:
                            if temp_cut_off[1] == 'ev':
                                param['cut_off_energy'] = f90_float_parse(temp_cut_off[0])
                            elif db:
                                raise RuntimeError('cut_off_energy units must be eV or blank in db mode, not {}'
                                                   .format(temp_cut_off[1]))
                            else:
                                param['cut_off_energy'] = '{} {}'.format(temp_cut_off[0], temp_cut_off[1])
                        else:
                            param['cut_off_energy'] = f90_float_parse(temp_cut_off[0])
                    # standardize tasks that have alternative spellings
                    elif 'task' in line and 'spectral_task' not in line and 'magres_task' not in line:
                        if 'singlepoint' in param['task']:
                            param['task'] = 'singlepoint'
                        elif 'geometry' in param['task']:
                            param['task'] = 'geometryoptimization'
                    elif 'xc_functional' in line:
                        param['xc_functional'] = param['xc_functional'].upper()
                    elif 'spin' in line and 'polar' not in line and 'spin' in param:
                        param['spin'] = int(param['spin'])
                    elif 'perc_extra_bands' in line:
                        param['perc_extra_bands'] = f90_float_parse(param['perc_extra_bands'])
                    elif 'geom_force_tol' in line:
                        param['geom_force_tol'] = f90_float_parse(param['geom_force_tol'])
                    elif 'elec_energy_tol' in line:
                        temp = (param['elec_energy_tol'].lower().replace('ev', ''))
                        temp = temp.strip()
                        param['elec_energy_tol'] = f90_float_parse(temp)
                    break

    if unrecognised:
        raise RuntimeError('Found several unrecognised parameters: {}'.format(unrecognised))

    return param, True


@scraper_function
def castep2dict(seed, db=True, intermediates=False, **kwargs):
    """ From seed filename, create dict of the most relevant
    information about a calculation.

    Parameters:
        seed (str/list): filename or list of filenames of castep file(s)

    Keyword arguments:
        db (bool): whether to error on missing relaxation info
        intermediates (bool): instead of a single dict containing the relaxed structure
            return a list of snapshots found in .castep file

    Returns:
        (tuple): containing either dict/str containing data or error, and a bool stating
            if the scrape was successful.

    """
    if seed.endswith('.history'):
        seed = seed.replace('.history', '')
        ftype = 'history'
    elif seed.endswith('.history.gz'):
        seed = seed.replace('.history.gz', '')
        ftype = 'history.gz'
    else:
        seed = seed.replace('.castep', '')
        ftype = 'castep'

    seed = '{}.{}'.format(seed, ftype)

    castep = dict()
    # read .castep, .history or .history.gz file
    if ftype == 'history.gz':
        with gzip.open(seed, 'r') as f:
            flines = [line.decode('utf-8') for line in f.readlines()]
    else:
        try:
            with open(seed, 'r', encoding='utf-8') as f:
                flines = f.readlines()
        except Exception:
            with open(seed, 'r', encoding='latin1') as f:
                flines = f.readlines()
    # set source tag to castep file
    castep['source'] = []
    castep['source'].append(seed)
    # grab file owner
    try:
        castep['user'] = pwd.getpwuid(os.stat(seed).st_uid).pw_name
    except Exception:
        pass
    get_seed_metadata(castep, seed.replace('.' + ftype, ''))
    # wrangle castep file for parameters in 3 passes:
    # once forwards to get number and types of atoms
    _castep_scrape_atoms(flines, castep)
    # once backwards to get the final parameter set for the calculation
    _castep_scrape_final_parameters(flines, castep)

    # task specific options
    if db and 'geometry' not in castep['task']:
        raise RuntimeError('CASTEP file does not contain GO calculation')

    if not db and 'thermo' in castep['task'].lower():
        _castep_scrape_thermo_data(flines, castep)

    if not db and 'thermo' in castep['task'].lower() or 'phonon' in castep['task'].lower():
        _castep_scrape_phonon_frequencies(flines, castep)

    # only scrape snapshots/number of intermediates if requested
    try:
        snapshots, castep['geom_iter'] = _castep_scrape_all_snapshots(flines)
        if intermediates:
            castep['intermediates'] = snapshots
    except RuntimeError as exc:
        if intermediates:
            raise RuntimeError('Failed to scrape intermediates: {}'.format(exc))

    _castep_scrape_metadata(flines, castep)

    # once more forwards, from the final step, to get the final structure
    _castep_scrape_final_structure(flines, castep, db=db)

    # scrape any BEEF post-processing
    _castep_scrape_beef(flines, castep)

    # scrape any AJM group-specific devel codes
    _castep_scrape_devel_code(flines, castep)

    if 'positions_frac' not in castep or not castep['positions_frac']:
        raise ComputationError('Could not find positions')

    # unfortunately CASTEP does not write forces when there is only one atom
    if 'forces' not in castep and castep['num_atoms'] == 1 and 'geometry' in castep['task']:
        castep['forces'] = [[0, 0, 0]]

    # finally check for pseudopotential files if OTF is present in species_pot
    if db:
        for species in castep['species_pot']:
            if 'OTF' in castep['species_pot'][species].upper():
                pspot_seed = ''
                for directory in seed.split('/')[:-1]:
                    pspot_seed += directory + '/'
                # glob for all .usp files with format species_*OTF.usp
                pspot_seed += castep['species_pot'][species]
                for globbed in glob.glob(pspot_seed):
                    if os.path.isfile(globbed):
                        castep['species_pot'].update(usp2dict(globbed))

    # check that any optimized results were saved and raise errors if not
    if not castep.get('optimised'):
        castep['optimised'] = False
        if db:
            # if importing to db, skip unconverged structure
            raise DFTError('CASTEP GO failed to converge.')

    return castep, True


@scraper_function
def bands2dict(seed, **kwargs):
    """ Parse a CASTEP bands file into a dictionary, which can be used as input to
    an :obj:`matador.orm.spectral.ElectronicDispersion` object.

    Parameters:
        seed (str/list): filename of list of filenames to be scraped.

    Returns:
        (tuple): containing either dict/str containing data or error, and a bool stating
            if the scrape was successful.

    """
    from matador.utils.chem_utils import HARTREE_TO_EV, BOHR_TO_ANGSTROM

    bs = dict()
    with open(seed, 'r') as f:
        # read whole file into RAM, typically ~ 1 MB
        flines = f.readlines()
    header = flines[:9]
    data = flines[9:]

    bs['source'] = [seed]

    bs['num_kpoints'] = int(header[0].split()[-1])
    bs['num_spins'] = int(header[1].split()[-1])
    bs['num_electrons'] = f90_float_parse(header[2].split()[-1])
    bs['num_bands'] = int(header[3].split()[-1])

    bs['spin_fermi_energy_Ha'] = [f90_float_parse(val) for val in header[4].split()[-bs['num_spins']:]]
    bs['spin_fermi_energy'] = [HARTREE_TO_EV * val for val in bs['spin_fermi_energy_Ha']]
    bs['fermi_energy'] = np.mean(bs['spin_fermi_energy'])

    bs['lattice_cart'] = []
    for i in range(3):
        bs['lattice_cart'].append([BOHR_TO_ANGSTROM * f90_float_parse(elem) for elem in header[6 + i].split()])
    bs['kpoint_path'] = np.zeros((bs['num_kpoints'], 3))
    bs['kpoint_weights'] = np.zeros((bs['num_kpoints']))
    bs['eigenvalues_k_s'] = np.empty((bs['num_spins'], bs['num_bands'], bs['num_kpoints']))

    for nk in range(bs['num_kpoints']):
        kpt_ind = nk * (bs['num_spins'] * bs['num_bands'] + bs['num_spins'] + 1)
        bs['kpoint_path'][int(data[kpt_ind].split()[1]) - 1] = (
            np.asarray([f90_float_parse(elem) for elem in data[kpt_ind].split()[-4:-1]])
        )
        bs['kpoint_weights'][int(data[kpt_ind].split()[1]) - 1] = f90_float_parse(data[kpt_ind].split()[-1])
        for nb in range(bs['num_bands']):
            for ns in range(bs['num_spins']):
                line_number = kpt_ind + 2 + ns + (ns * bs['num_bands']) + nb
                bs['eigenvalues_k_s'][ns][nb][int(data[kpt_ind].split()[1]) - 1] = (
                    f90_float_parse(data[line_number].strip()))

    bs['eigenvalues_k_s'] *= HARTREE_TO_EV
    bs['eigs_s_k'] = bs['eigenvalues_k_s']

    return bs, True


@scraper_function
def arbitrary2dict(seed, **kwargs):
    """ Read arbitrary CASTEP-style input files into
    a dictionary.

    Parameters:
        seed (str/list): filename or list of filenames.

    Returns:
        (tuple): containing either dict/str containing data or error, and a bool stating
            if the scrape was successful.

    """
    with open(seed, 'r') as f:
        flines = f.readlines()
    splitters = [':', '=', '\t', ' ']
    comment_delims = ['!', '%', '#', '/*', '*/']
    result = {}
    result['source'] = [seed]
    for line in flines:
        comment = False
        for delim in comment_delims:
            if line.strip().startswith(delim):
                comment = True
                break
        if not comment:
            for splitter in splitters:
                if splitter in line:
                    keyword = line.split(splitter)[0].strip()
                    value = line.split(splitter)[-1].strip()
                    result[keyword.lower()] = value
                    break

    return result, True


@scraper_function
def optados2dict(seed, **kwargs):
    """ Scrape optados output file (*.*.dat) or (*.pdos.*.dat)
    for DOS, projectors and projected DOS/dispersion.

    Parameters:
        seed (str/list): optados filename or list of filenames.

    Returns:
        (tuple): containing either dict/str containing data or error, and a bool stating
            if the scrape was successful.

    """
    optados = dict()
    is_pdos = False
    is_pdis = False
    is_spin_dos = False
    dos_unit_label = None
    with open(seed, 'r') as f:
        flines = f.readlines()

    header = []
    header_ind = 0
    for ind, line in enumerate(flines):
        header_ind = ind
        if not line.strip().startswith('#') or 'K-point' in line:
            break
        if 'Partial' in line:
            is_pdos = True
        elif 'spin' in line:
            is_spin_dos = True
        elif 'Projected Dispersion' in line:
            is_pdis = True
        elif 'DOS (' in line and 'Integrated' not in line:
            dos_unit_label = ' '.join(line.split()[2:])
        else:
            header.append(line)

    flines = flines[header_ind:]

    if not is_pdis:
        data = np.loadtxt(seed, comments='#')
        optados['energies'] = data[:, 0]

    if is_pdos or is_pdis:
        optados['projectors'] = _optados_get_projector_labels(header)
        optados['num_projectors'] = len(optados['projectors'])

    if dos_unit_label is not None:
        optados['dos_unit_label'] = dos_unit_label

    if is_pdos:
        # get pdos values
        optados['pdos'] = dict()
        optados['sum_pdos'] = np.zeros_like(data[:, 0])
        for i, projector in enumerate(optados['projectors']):
            # optados spin-down projectors are negative, unfortunately
            if 'down' in projector:
                optados['pdos'][projector] = -data[:, i + 1]
            else:
                optados['pdos'][projector] = data[:, i + 1]
            optados['sum_pdos'] += data[:, i + 1]

    elif is_spin_dos:
        optados['spin_dos'] = dict()
        optados['spin_dos']['up'] = data[:, 1]
        optados['spin_dos']['down'] = -data[:, 2]
        optados['dos'] = np.abs(optados['spin_dos']['up']) + np.abs(optados['spin_dos']['down'])

    elif is_pdis:
        optados['projector_weights'] = []
        optados['kpoints'] = []
        optados['eigenvalues'] = []
        # get kpoints and count number of bands
        kpt_ind = -1
        for i, line in enumerate(flines):
            if 'K-point' in line:
                optados['kpoints'].append([f90_float_parse(val) for val in line.split()[-3:]])
                if kpt_ind == -1:
                    kpt_ind = i
                else:
                    if not optados.get('num_bands'):
                        optados['num_bands'] = i - kpt_ind - 1

        optados['num_kpoints'] = len(optados['kpoints'])

        for nk in range(optados['num_kpoints']):
            eigs = []
            pdis = []
            for nb in range(0, optados['num_bands']):
                eigs.append(f90_float_parse(flines[nk*(optados['num_bands']+1) + nb + 1].split()[0]))
                pdis.append([f90_float_parse(val) for val in flines[nk*(optados['num_bands']+1) + 1 + nb].split()[1:]])
            optados['eigenvalues'].append(eigs)
            optados['projector_weights'].append(pdis)

        optados['projector_weights'] = np.asarray(optados['projector_weights'])

    else:
        optados['dos'] = data[:, 1]

    return optados, True


@scraper_function
def phonon2dict(seed, **kwargs):
    """ Parse a CASTEP phonon file into a dictionary.

    Parameters:
        seed (str/list): phonon filename or list of filenames.

    Returns:
        (tuple): containing either dict/str containing data or error, and a bool stating
            if the scrape was successful.

    """

    with open(seed, 'r') as f:
        # read whole file into RAM, typically <~ 1 MB
        flines = f.readlines()

    ph = dict()

    if 'phonon_dos' in seed:
        seed = seed.replace('.phonon_dos', '')
        ph['source'] = [seed + '.phonon_dos']
    else:
        seed = seed.replace('.phonon', '')
        ph['source'] = [seed + '.phonon']

    dos_present = False
    data_start = 0

    for line_no, line in enumerate(flines):
        line = line.lower()
        if 'number of ions' in line:
            ph['num_atoms'] = int(line.split()[-1])
        elif 'number of branches' in line:
            ph['num_modes'] = int(line.split()[-1])
            ph['num_branches'] = ph['num_modes']
        elif 'number of wavevectors' in line:
            ph['num_kpoints'] = int(line.split()[-1])
        elif 'frequencies in' in line:
            ph['freq_unit'] = line.split()[-1]
        elif 'unit cell vectors' in line:
            ph['lattice_cart'] = []
            for i in range(3):
                ph['lattice_cart'].append([f90_float_parse(elem) for elem in flines[line_no + i + 1].split()])
        elif 'fractional co-ordinates' in line:
            ph['positions_frac'] = []
            ph['atom_types'] = []
            ph['atom_masses'] = []
            i = 1
            while 'END header' not in flines[line_no + i]:
                ph['positions_frac'].append([f90_float_parse(elem) for elem in flines[line_no + i].split()[1:4]])
                ph['atom_types'].append(flines[line_no + i].split()[-2])
                ph['atom_masses'].append(f90_float_parse(flines[line_no + i].split()[-1]))
                i += 1
        elif 'end header' in line:
            data_start = line_no + 1
        elif 'begin dos' in line:
            dos_present = True
            projector_labels = flines[line_no].split()[5:]
            projector_labels = [(label, None) for label in projector_labels]
            begin_dos = line_no + 1

        elif 'q-pt' in line:
            last_qpt_ind = int(line.split()[1])

    if dos_present:
        # extra header line with GRADIENTS written when dos is present
        data_start += 1
        # no eigenvectors written when dos is present
        line_offset = ph['num_modes'] + 1
    else:
        line_offset = ph['num_modes'] * (ph['num_atoms'] + 1) + 3

    data = flines[data_start:]

    ph['phonon_kpoint_list'] = []
    if 'num_kpoints' not in ph:
        ph['num_kpoints'] = last_qpt_ind
    ph['eigenvalues_q'] = np.zeros((1, ph['num_modes'], ph['num_kpoints']))
    raman_intensity = np.zeros_like(ph['eigenvalues_q'])
    infrared_intensity = np.zeros_like(ph['eigenvalues_q'])
    raman = False
    ir = False
    for qind in range(ph['num_kpoints']):
        ph['phonon_kpoint_list'].append([f90_float_parse(elem) for elem in data[qind * line_offset].split()[2:]])
        for i in range(1, ph['num_modes'] + 1):
            line_split = data[qind * line_offset + i].split()
            ph['eigenvalues_q'][0][i - 1][qind] = f90_float_parse(line_split[1])
            if len(line_split) > 2:
                infrared_intensity[0][i - 1][qind] = f90_float_parse(line_split[2])
                ir = True
            if len(line_split) > 3:
                raman_intensity[0][i - 1][qind] = f90_float_parse(line_split[3])
                raman = True

    if ir:
        ph['infrared_intensity'] = infrared_intensity
    if raman:
        ph['raman_intensity'] = raman_intensity

    if dos_present:
        # remove header and "END"
        flines = flines[begin_dos:-1]
        raw_data = np.genfromtxt(flines)
        ph['energies'] = raw_data[:, 0] * INVERSE_CM_TO_EV
        ph['dos'] = raw_data[:, 1]
        ph['pdos'] = dict()

        for i, label in enumerate(projector_labels):
            ph['pdos'][label] = raw_data[:, i + 2]

    ph['kpoint_path'] = np.asarray([qpt[0:3] for qpt in ph['phonon_kpoint_list']])
    ph['kpoint_weights'] = [qpt[3] for qpt in ph['phonon_kpoint_list']]
    ph['eigenvalues_q'] *= INVERSE_CM_TO_EV
    ph['softest_mode_freq'] = np.min(ph['eigenvalues_q'])
    ph['eigs_q'] = ph['eigenvalues_q']

    return ph, True


@scraper_function
def phonon_dos2dict(*args, **kwargs):
    """ Wrapper for old phonon DOS scraper, which has since been merged
    with `phonon2dict`. Note that this function still has a different
    effect to `phonon2dict` when `as_model` is used as the results will
    be cast into a :class:`VibrationalDOS` object.

    """
    return phonon2dict(*args, no_wrap=True, **kwargs)


def usp2dict(seed, **kwargs):
    """ Extract pseudopotential string from a CASTEP
    OTF .USP file.

    Parameters:
        seed (str/list): filename of usp file, or list of filenames.

    Returns:
        dict: partial species_pot dict from usp file.

    """
    species_pot = dict()
    with open(seed, 'r') as f:
        flines = f.readlines()
        for line_no, line in enumerate(flines):
            if 'Pseudopotential Report' in line:
                i = 0
                while i + line_no < len(flines) - 3:
                    if 'Pseudopotential Report' in flines[line_no + i]:
                        i += 2
                        elem = flines[line_no + i].split(':')[1].split()[0]
                    elif 'core correction' in flines[line_no + i]:
                        i += 2
                        species_pot[elem] = flines[line_no + i].strip().split()[1]
                        # check next line for wrapped definition
                        if flines[line_no + i + 1].strip().startswith('--------'):
                            break
                        else:
                            species_pot[elem] += flines[line_no + i + 1].strip().split()[1]
                            break
                    i += 1
    species_pot[elem] = species_pot[elem].replace('"', '')
    species_pot[elem] = species_pot[elem].replace('[]', '')
    return species_pot


def _castep_scrape_thermo_data(flines, castep):
    """ Scrape the data from a CASTEP Thermodynamics claculation.

    Note:
        This only scrapes from Thermodynamics section currently,
        NOT Atomic Displacement Parameters

    Parameters:
        flines (list): list of lines contained in file
        castep (dict): dictionary to update with data

    Returns:
        dict: dictionary updated with scraped thermodynamics data
              in format doc['energy_type'] = {300:-0.01,200:-0.05}

    """
    for line_no, line in enumerate(flines):
        if 'Number of temperature values' in line:
            castep['thermo_num_temp_vals'] = int(line.split(':')[-1].strip())
        elif 'Initial temperature' in line:
            castep['thermo_temp_init'] = f90_float_parse(line.split(':')[1].strip().split(' ')[0])
        elif 'Final temperature' in line:
            castep['thermo_temp_final'] = f90_float_parse(line.split(':')[1].strip().split(' ')[0])
        elif 'Spacing between temperature values' in line:
            castep['thermo_temp_spacing'] = f90_float_parse(line.split(':')[1].strip().split(' ')[0])
        elif 'Zero-point energy' in line:
            castep['thermo_zero_point_energy'] = f90_float_parse(line.split("=")[1].strip().split(' ')[0])
        elif 'T(K)' and 'E(eV)' in line:
            castep['thermo_temps'] = []  # temperatures calculation was done at
            castep['thermo_enthalpy'] = {}  # enthalpy E(eV)
            castep['thermo_free_energy'] = {}  # free energy F(eV)
            castep['thermo_entropy'] = {}  # entropy S(J/mol/K)
            castep['thermo_heat_cap'] = {}  # heat capacity Cv(J/mol/K)
            i = 2
            while True:
                if not flines[line_no + i + 1].strip():
                    break
                else:
                    temp_line = flines[line_no + i].split()
                    castep['thermo_temps'].append(f90_float_parse(temp_line[0]))
                    castep['thermo_enthalpy'][f90_float_parse(temp_line[0])] = f90_float_parse(temp_line[1])
                    castep['thermo_free_energy'][f90_float_parse(temp_line[0])] = f90_float_parse(temp_line[2])
                    castep['thermo_entropy'][f90_float_parse(temp_line[0])] = f90_float_parse(temp_line[3])
                    castep['thermo_heat_cap'][f90_float_parse(temp_line[0])] = f90_float_parse(temp_line[4])
                i += 1


def _castep_scrape_phonon_frequencies(flines, castep):
    """ Iterate through flines to scrape the phonon frequencies
    from this CASTEP calculation. Will only scrape the *final* set
    of frequencies in the CASTEP file, ignoring any others.

    """
    phonons = {}

    phonons['phonon_fine_kpoint_list'] = []
    phonons['phonon_fine_kpoint_weights'] = []
    phonons['eigs_q'] = []

    # first find the last block of frequencies
    for line_no, line in enumerate(flines):
        if "Performing frequency calculation at " in line:
            start_line_no = line_no + 2

    q_pt_ind = 0
    for line_no, line in enumerate(flines[start_line_no:]):
        if '============' in line:
            break
        if 'q-pt=' in line:
            q_pt = [f90_float_parse(val) for val in line.split('(')[-1].split(')')[0].split()]
            phonons['phonon_fine_kpoint_list'].append(q_pt)
            phonons['phonon_fine_kpoint_weights'].append(f90_float_parse(line.split()[-2]))
            phonons['eigs_q'].append([])

            for _, freq_line in enumerate(flines[start_line_no:][line_no+6:]):
                if '.........................' in freq_line:
                    break
                phonons['eigs_q'][q_pt_ind].append(f90_float_parse(freq_line.split()[2]))
            q_pt_ind += 1

    phonons['num_modes'] = len(phonons['eigs_q'][0])
    phonons['eigs_q'] = np.asarray(phonons['eigs_q']).T
    phonons['eigs_q'] = INVERSE_CM_TO_EV * phonons['eigs_q'].reshape(1, *np.shape(phonons['eigs_q']))
    phonons['kpoint_path'] = phonons['phonon_fine_kpoint_list']
    phonons['kpoint_weights'] = phonons['phonon_fine_kpoint_weights']
    phonons['num_kpoints'] = len(phonons['phonon_fine_kpoint_list'])
    phonons['num_qpoints'] = len(phonons['phonon_fine_kpoint_list'])

    for key in phonons:
        castep[key] = phonons[key]


def _castep_scrape_atoms(flines, castep):
    """ Iterate forwards through flines to scrape atomic types and
    initial positions, and get a preliminary lattice.

    Parameters:
        flines (list): list of lines in file
        castep (dict): dictionary to update with scraped data

    Returns:
        dict: dictionary updated with scraped data

    """
    for line_no, line in enumerate(flines):
        if 'Real Lattice' in line:
            castep['lattice_cart'] = []
            i = 1
            while True:
                if not flines[line_no + i].strip():
                    break
                else:
                    temp_line = flines[line_no + i].split()[0:3]
                    castep['lattice_cart'].append(list(map(f90_float_parse, temp_line)))
                i += 1
        if 'Lattice parameters' in line:
            castep['lattice_abc'] = []
            i = 1
            castep['lattice_abc'].append(
                list(map(f90_float_parse,
                         [flines[line_no+i].split('=')[1].strip().split(' ')[0],
                          flines[line_no+i+1].split('=')[1].strip().split(' ')[0],
                          flines[line_no+i+2].split('=')[1].strip().split(' ')[0]])))
            castep['lattice_abc'].append(
                list(map(f90_float_parse,
                         [flines[line_no+i].split('=')[-1].strip(),
                          flines[line_no+i+1].split('=')[-1].strip(),
                          flines[line_no+i+2].split('=')[-1].strip()])))
        if 'Current cell volume' in line:
            castep['cell_volume'] = f90_float_parse(line.split('=')[1].split()[0].strip())
        if 'atom types' not in castep and 'Cell Contents' in line:
            castep['atom_types'] = []
            castep['positions_frac'] = []
            i = 1
            atoms = False
            while True:
                if atoms:
                    if 'xxxxxxxxx' in flines[line_no + i]:
                        atoms = False
                        break
                    else:
                        castep['atom_types'].append(flines[line_no + i].split()[1])
                        castep['positions_frac'].append(list(map(f90_float_parse, (flines[line_no + i].split()[3:6]))))
                if 'x------' in flines[line_no + i]:
                    atoms = True
                i += 1
            for ind, pos in enumerate(castep['positions_frac']):
                for k in range(3):
                    if pos[k] > 1 or pos[k] < 0:
                        castep['positions_frac'][ind][k] %= 1
            castep['num_atoms'] = len(castep['atom_types'])
            castep['stoichiometry'] = get_stoich(castep['atom_types'])
            castep['num_fu'] = castep['num_atoms'] / sum([elem[1] for elem in castep['stoichiometry']])
            break
    else:
        raise ComputationError('Unable to find atoms in CASTEP file.')


def _castep_scrape_final_parameters(flines, castep):
    """ Scrape the DFT parameters from a CASTEP file, using those listed
    last in the file (i.e. those used to make the final structure).

    Parameters:
        flines (list): list of lines in the file
        castep (dict): dictionary in which to put scraped data

    Returns:
        dict: dictionary updated with scraped data

    """

    pspot_report_dict = dict()
    for line_no, line in enumerate(reversed(flines)):
        line_no = len(flines) - 1 - line_no
        if 'task' not in castep and 'type of calculation' in line:
            castep['task'] = line.split(':')[-1].strip().replace(" ", "")
        elif 'xc_functional' not in castep and 'functional' in line:
            # convert from .castep file xc_functional to param style
            xc_string = line.split(':')[-1].strip()
            if 'Local Density Approximation' in xc_string:
                castep['xc_functional'] = 'LDA'
            elif 'Perdew Burke Ernzerhof' in xc_string:
                castep['xc_functional'] = 'PBE'
            elif 'PBE for solids' in xc_string:
                castep['xc_functional'] = 'PBESol'
            elif 'hybrid B3LYP' in xc_string:
                castep['xc_functional'] = 'B3LYP'
            elif 'hybrid HSE03' in xc_string:
                castep['xc_functional'] = 'HSE03'
            elif 'hybrid HSE06' in xc_string:
                castep['xc_functional'] = 'HSE06'
            elif 'RSCAN' in xc_string:
                castep['xc_functional'] = 'RSCAN'
            elif 'Screened Hartree-Fock' in xc_string:
                castep['xc_functional'] = 'SHF-LDA'
            else:
                castep['xc_functional'] = xc_string.split(':')[-1]
                warnings.warn('Unrecognised functional {xc}: scraping as {xc}.'
                              'This may lead to incompatible param files.'
                              .format(xc=castep['xc_functional']))
        elif 'cut_off_energy' not in castep and 'plane wave basis set' in line:
            castep['cut_off_energy'] = f90_float_parse(line.split(':')[-1].split()[0])
        elif 'finite_basis_corr' not in castep and 'finite basis set correction  ' in line:
            castep['finite_basis_corr'] = line.split(':')[-1].strip()
        elif 'kpoints_mp_grid' not in castep and 'MP grid size for SCF' in line:
            castep['kpoints_mp_grid'] = list(map(int, list(line.split('is')[-1].split())))
        elif 'num_kpoints' not in castep and 'Number of kpoints used' in line:
            castep['num_kpoints'] = int(line.split()[-1])
        elif 'geom_force_tol' not in castep and 'max ionic |force| tolerance' in line:
            castep['geom_force_tol'] = f90_float_parse(line.split()[-2])
        elif 'elec_energy_tol' not in castep and 'total energy / atom convergence tol' in line:
            castep['elec_energy_tol'] = f90_float_parse(line.split()[-2])
        elif 'sedc_apply' not in castep and \
                'DFT+D: Semi-empirical dispersion correction    : on' in line:
            castep['sedc_apply'] = True
            castep['sedc_scheme'] = flines[line_no + 1].split(':')[1].split()[0]
        elif 'space_group' not in castep and 'Space group of crystal' in line:
            space_group = line.split(':')[-1].split(',')[0].strip().replace(" ", "")
            if space_group:
                castep['space_group'] = space_group
        elif 'nelectrons' not in castep and 'number of  electrons' in line:
            castep['nelectrons'] = f90_float_parse(line.split(':')[-1])
        elif 'nelectrons' not in castep and 'number of bands' in line:
            castep['nbands'] = int(line.split(':')[-1])
        elif 'Cell constraints are' in line and 'cell_constraints' not in castep:
            castep['cell_constraints'] = [int(val) for val in line.split(':')[-1].split()]
            if castep['cell_constraints'] == [1, 2, 3, 4, 5, 6]:
                del castep['cell_constraints']
            elif all(val == 0 for val in castep['cell_constraints']):
                castep['fix_all_cell'] = True
                del castep['cell_constraints']
            if 'cell_constraints' in castep:
                if len(castep['cell_constraints']) != 6:
                    raise RuntimeError('Unable to read cell constraints block.')
                castep['cell_constraints'] = [castep['cell_constraints'][0:3], castep['cell_constraints'][3:]]
        elif 'external_pressure' not in castep and 'External pressure/stress' in line:
            try:
                castep['external_pressure'] = []
                for i in range(3):
                    parsed_line = list(map(f90_float_parse, flines[line_no + i + 1].split()))
                    if len(parsed_line) != 3:
                        parsed_line = i * [0.0] + parsed_line
                    castep['external_pressure'].append(parsed_line)
            except ValueError:
                castep['external_pressure'] = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
        elif 'spin_polarized' not in castep and 'treating system as spin-polarized' in line:
            castep['spin_polarized'] = True
        elif 'hubbard_u' not in castep and 'Hubbard U values are eV' in line:
            castep['hubbard_u'] = defaultdict(list)
            i = 5
            while i < castep['num_atoms']:
                line = flines[line_no + i].strip()
                atom = line.split()[0].replace('|', '')
                shifts = list(map(f90_float_parse, line.split()[-5:-1]))
                for ind, shift in enumerate(shifts):
                    if shift != 0:
                        if atom not in castep['hubbard_u']:
                            castep['hubbard_u'][atom] = dict()
                        if ind == 0:
                            orbital = 's'
                        elif ind == 1:
                            orbital = 'p'
                        elif ind == 2:
                            orbital = 'd'
                        elif ind == 3:
                            orbital = 'f'
                        castep['hubbard_u'][atom][orbital] = shift
                i += 1
        elif 'Pseudopotential Report' in line:
            if 'species_pot' not in castep:
                castep['species_pot'] = dict()
            i = 0
            while i + line_no < len(flines) - 3:
                if 'Pseudopotential Report' in flines[line_no + i]:
                    i += 2
                    elem = flines[line_no + i].split(':')[1].split()[0]

                elif 'core correction' in flines[line_no + i]:
                    i += 2
                    if not pspot_report_dict.get(elem):
                        castep['species_pot'][elem] = flines[line_no + i].split('"')[1].replace('[]', '')
                        pspot_report_dict[elem] = True
                i += 1
        elif 'species_pot' not in castep and 'Files used for pseudopotentials' in line:
            if 'species_pot' not in castep:
                castep['species_pot'] = dict()
            i = 1
            while True:
                if not flines[line_no + i].strip():
                    break
                else:
                    elem = flines[line_no + i].split()[0].strip()
                    if not pspot_report_dict.get(elem):
                        castep['species_pot'][elem] = flines[line_no + i].split()[1].split('/')[-1]
                        if castep['species_pot'][elem] == 'Pseudopotential':
                            castep['species_pot'][elem] = flines[line_no + i].split()[0].strip()
                            castep['species_pot'][elem] += '_OTF.usp'
                        pspot_report_dict[elem] = False
                i += 1
    # write zero pressure if not found in file
    if 'external_pressure' not in castep:
        castep['external_pressure'] = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
    if 'spin_polarized' not in castep:
        castep['spin_polarized'] = False


def _castep_scrape_final_structure(flines, castep, db=True):
    """ Scrape final structure from CASTEP file.

    Parameters:
        flines (list): list of lines contained in file
        castep (dict): dictionary to update with data

    Keyword arguments:
        db (bool): whether to enforce database style, e.g. geometry optimisation only

    Returns:
        dict: dictionary updated with scraped data

    """
    if 'task' in castep and castep['task'].strip() == 'geometryoptimization':
        castep['optimised'] = False
        finish_line, castep['optimised'] = _castep_find_final_structure(flines)
    elif not db:
        finish_line = 0

    final_flines = flines[finish_line + 1:]
    for line_no, line in enumerate(final_flines):
        if 'Real Lattice' in line:
            castep['lattice_cart'] = []
            i = 1
            while i < 4:
                if not final_flines[line_no + i].strip():
                    break
                else:
                    temp_line = final_flines[line_no + i].split()[0:3]
                    castep['lattice_cart'].append(list(map(f90_float_parse, temp_line)))
                i += 1
        if 'Lattice parameters' in line:
            castep['lattice_abc'] = []
            i = 1
            castep['lattice_abc'].append(
                list(map(f90_float_parse,
                         [final_flines[line_no+i].split('=')[1].strip().split(' ')[0],
                          final_flines[line_no+i+1].split('=')[1].strip().split(' ')[0],
                          final_flines[line_no+i+2].split('=')[1].strip().split(' ')[0]])))
            castep['lattice_abc'].append(
                list(map(f90_float_parse,
                         [final_flines[line_no+i].split('=')[-1].strip(),
                          final_flines[line_no+i+1].split('=')[-1].strip(),
                          final_flines[line_no+i+2].split('=')[-1].strip()])))
        elif 'Current cell volume' in line:
            castep['cell_volume'] = f90_float_parse(line.split('=')[1].split()[0].strip())
        elif 'Cell Contents' in line:
            castep['positions_frac'] = []
            i = 1
            atoms = False
            while True:
                if atoms:
                    if 'xxxxxxxxx' in final_flines[line_no + i]:
                        atoms = False
                        break
                    else:
                        temp_frac = final_flines[line_no + i].split()[3:6]
                        castep['positions_frac'].append(list(map(f90_float_parse, temp_frac)))
                if 'x------' in final_flines[line_no + i]:
                    atoms = True
                i += 1
        # don't check if final_energy exists, as this will update for each GO step
        elif 'Final energy =' in line or 'Final energy, E' in line:
            castep['total_energy'] = f90_float_parse(line.split('=')[1].split()[0])
            castep['total_energy_per_atom'] = castep['total_energy'] / castep['num_atoms']
        elif 'Final free energy' in line:
            castep['smeared_free_energy'] = f90_float_parse(line.split('=')[1].split()[0])
            castep['smeared_free_energy_per_atom'] = castep['smeared_free_energy'] / castep['num_atoms']
        elif '0K energy' in line:
            castep['0K_energy'] = f90_float_parse(line.split('=')[1].split()[0])
            castep['0K_energy_per_atom'] = castep['0K_energy'] / castep['num_atoms']
        elif ' Forces **' in line:
            castep['forces'] = []
            i = 1
            forces = False
            while True:
                if forces:
                    if '*' in final_flines[line_no + i].split()[1]:
                        forces = False
                        break
                    else:
                        castep['forces'].append([])
                        for j in range(3):
                            temp = final_flines[line_no + i].replace('(cons\'d)', '')
                            castep['forces'][-1].append(f90_float_parse(temp.split()[3 + j]))
                elif 'x' in final_flines[line_no + i]:
                    i += 1  # skip next blank line
                    forces = True
                i += 1
            castep['max_force_on_atom'] = np.max(np.linalg.norm(castep['forces'], axis=-1))
        elif 'Stress Tensor' in line:
            i = 1
            while i < 20:
                if 'Cartesian components' in final_flines[line_no + i]:
                    castep['stress'] = []
                    for j in range(3):
                        castep['stress'].append(list(map(f90_float_parse, (final_flines[line_no + i + j + 4].split()[2:5]))))
                elif 'Pressure' in final_flines[line_no + i]:
                    try:
                        castep['pressure'] = f90_float_parse(final_flines[line_no + i].split()[-2])
                    except ValueError:
                        pass
                    break
                i += 1
        elif 'Integrated Spin Density' in line:
            castep['integrated_spin_density'] = f90_float_parse(line.split()[-2])
        elif 'Integrated |Spin Density|' in line:
            castep['integrated_mod_spin_density'] = f90_float_parse(line.split()[-2])
        elif 'Atomic Populations (Mulliken)' in line:
            # population format seems to change every CASTEP version...
            if float(castep.get('castep_version', 0.0)) >= 18:
                if castep['spin_polarized']:
                    castep['mulliken_spins'] = []
                    castep['mulliken_net_spin'] = 0.0
                    castep['mulliken_abs_spin'] = 0.0
                castep['mulliken_charges'] = []
                castep['mulliken_spins'] = []
                i = 0
                ind = 0
                while ind < len(castep['atom_types']):
                    if castep['spin_polarized']:
                        castep['mulliken_charges'].append(f90_float_parse(final_flines[line_no + i + 4].split()[-2]))
                        castep['mulliken_spins'].append(f90_float_parse(final_flines[line_no + i + 4].split()[-1]))
                        castep['mulliken_net_spin'] += castep['mulliken_spins'][-1]
                        castep['mulliken_abs_spin'] += abs(castep['mulliken_spins'][-1])
                        i += 2
                    else:
                        castep['mulliken_charges'].append(f90_float_parse(final_flines[line_no + i + 4].split()[-1]))
                        i += 1
                    ind += 1
        elif 'Final Enthalpy' in line:
            castep['enthalpy'] = f90_float_parse(line.split('=')[-1].split()[0])
            castep['enthalpy_per_atom'] = (f90_float_parse(line.split('=')[-1].split()[0]) / castep['num_atoms'])
        elif 'Final bulk modulus' in line:
            try:
                castep['bulk_modulus'] = f90_float_parse(line.split('=')[-1].split()[0])
            except ValueError:
                # the above will fail if bulk modulus was not printed (i.e. if it was unchanged)
                pass

        elif 'Chemical Shielding and Electric Field Gradient Tensors'.lower() in line.lower():
            i = 5
            castep['chemical_shifts'] = []
            while True:
                # break when the line containing just '=' is reached
                if len(flines[line_no + i].split()) == 1:
                    break
                castep['chemical_shifts'].append(flines[line_no + i].split()[3])
                i += 1
            if len(castep['chemical_shifts']) != len(castep['atom_types']):
                raise RuntimeError('Found fewer chemical shifts than atoms (or vice versa)!')

    # calculate kpoint spacing if not found
    if 'kpoints_mp_grid' in castep and 'kpoints_mp_spacing' not in castep and 'lattice_cart' in castep:
        castep['kpoints_mp_spacing'] = calc_mp_spacing(castep['lattice_cart'], castep['kpoints_mp_grid'], prec=4)
    return castep


def _castep_scrape_metadata(flines, castep):
    """ Scrape metadata from CASTEP file.

    Parameters:
        flines (list): list of lines contained in file
        castep (dict): dictionary to update with data

    Returns:
        dict: dictionary updated with scraped data

    """
    from time import strptime
    # computing metadata, i.e. parallelism, time, memory, version
    if 'total_time_secs' not in castep:
        castep['total_time_secs'] = 0
    if 'total_time_hrs' not in castep:
        castep['total_time_hrs'] = 0

    for ind, line in enumerate(flines):
        if 'Release CASTEP version' in line:
            castep['castep_version'] = line.replace('|', '').split()[-1]
        if 'Run started:' in line:
            try:
                year = line.split()[5]
                month = str(strptime(line.split()[4], '%b').tm_mon)
                day = line.split()[3]
                castep['date'] = day + '-' + month + '-' + year
            except (IndexError, ValueError):
                castep['date'] = 'unknown'
        elif 'compiled for' in line.lower():
            try:
                castep['_compiler_architecture'] = line.split()[2]
            except IndexError:
                castep['_compiler_architecture'] = 'unknown'
        elif 'from code version' in line.lower():
            try:
                castep['_castep_commit'] = ' '.join(flines[ind:ind+2]).split()[3]
            except (IndexError, ValueError):
                castep['_castep_commit'] = 'unknown'
        elif 'Total time' in line and 'matrix elements' not in line:
            try:
                time = f90_float_parse(line.split()[-2])
                castep['total_time_secs'] += time
                castep['total_time_hrs'] += time / 3600
            except (IndexError, ValueError):
                castep['final_calculation_time_secs'] = 0
        elif 'Calculation only took' in line:
            castep['_time_estimated'] = f90_float_parse(line.split()[4])
        elif 'Peak Memory Use' in line:
            try:
                castep['peak_mem_MB'] = int(f90_float_parse(line.split()[-2]) / 1024)
            except (IndexError, ValueError):
                castep['peak_mem_MB'] = -1
        elif 'total storage required per process' in line:
            try:
                castep['estimated_mem_per_process_MB'] = f90_float_parse(line.split()[-5])
            except (IndexError, ValueError):
                castep['estimated_mem_per_process_MB'] = 0
        elif 'Calculation parallelised over' in line:
            try:
                castep['num_mpi_processes'] = int(f90_float_parse(line.split()[3]))
            except Exception:
                castep['num_mpi_processes'] = 1
        elif 'Calculation not parall' in line:
            castep['num_mpi_processes'] = 1

        if 'num_mpi_processes' in castep and 'estimated_mem_per_process_MB' in castep:
            castep['estimated_mem_MB'] = castep['estimated_mem_per_process_MB'] * castep['num_mpi_processes']


def _castep_find_final_structure(flines):
    """ Search for info on final structure in .castep file.

    Parameters:
        flines (list): list of lines in file.

    Returns:
        int: line number in file where total energy of final structure is printed.

    """
    optimised = False
    finish_line = 0
    success_string = 'Geometry optimization completed successfully'
    failure_string = 'Geometry optimization failed to converge after'
    annoying_string = 'WARNING - there is nothing to optimise - skipping relaxation'
    # look for final "success/failure" string in file for geometry optimisation
    for line_no, line in enumerate(reversed(flines)):
        if success_string in line:
            finish_line = len(flines) - line_no
            optimised = True
            break
        if annoying_string in line:
            finish_line = len(flines) - line_no
            optimised = True
            break
        if failure_string in line:
            finish_line = len(flines) - line_no
            optimised = False
            break

    # now wind back to get final total energies and non-symmetrised forces
    for count, line in enumerate(reversed(flines[:finish_line])):
        if 'Final energy, E' in line or 'Final energy =' in line:
            finish_line -= count + 2
            break

    return finish_line, optimised


def _castep_scrape_all_snapshots(flines):
    """ Scrape all intermediate structures from a CASTEP file, both
    geometry optimisation snapshots, and repeated SCF calculations.

    Parameters:
        flines (list): list of lines of file.

    Returns:
        :obj:`list` of :obj:`dict`: list of dictionaries containing
            intermediate snapshots.
        int: number of completed geometry optimisation steps.

    """
    intermediates = []
    num_opt_steps = 0
    snapshot = dict()
    for line_no, line in enumerate(flines):
        try:
            # use the "Real Lattice" line as the start of a new snapshot / end of old one
            if 'Real Lattice' in line:
                # add the last snapshot only if it isn't a repeat
                if 'total_energy' in snapshot:
                    # if positions frac didn't change (and thus weren't printed, use the last value)
                    if 'positions_frac' not in snapshot:
                        snapshot['positions_frac'] = intermediates[-1]['positions_frac']
                        snapshot['atom_types'] = intermediates[-1]['atom_types']
                        snapshot['num_atoms'] = len(snapshot['positions_frac'])
                    snapshot['smeared_free_energy_per_atom'] = snapshot['smeared_free_energy'] / snapshot['num_atoms']
                    snapshot['total_energy_per_atom'] = snapshot['total_energy'] / snapshot['num_atoms']
                    snapshot['0K_energy_per_atom'] = snapshot['0K_energy'] / snapshot['num_atoms']
                    # handle single atom forces edge-case
                    if snapshot['num_atoms'] == 1:
                        snapshot['forces'] = [[0, 0, 0]]
                    if not intermediates:
                        intermediates.append(snapshot)
                    elif (snapshot['lattice_cart'] != intermediates[-1]['lattice_cart'] and
                          snapshot['total_energy'] != intermediates[-1]['total_energy']):
                        intermediates.append(snapshot)

                snapshot = dict()
                snapshot['lattice_cart'] = []
                i = 1
                while True:
                    if not flines[line_no + i].strip():
                        break
                    else:
                        temp_line = flines[line_no + i].split()[0:3]
                        snapshot['lattice_cart'].append(list(map(f90_float_parse, temp_line)))
                    i += 1

            elif 'Lattice parameters' in line:
                snapshot['lattice_abc'] = []
                i = 1
                snapshot['lattice_abc'].append(
                    list(map(f90_float_parse,
                             [flines[line_no+i].split('=')[1].strip().split(' ')[0],
                              flines[line_no+i+1].split('=')[1].strip().split(' ')[0],
                              flines[line_no+i+2].split('=')[1].strip().split(' ')[0]])))
                snapshot['lattice_abc'].append(
                    list(map(f90_float_parse,
                             [flines[line_no+i].split('=')[-1].strip(),
                              flines[line_no+i+1].split('=')[-1].strip(),
                              flines[line_no+i+2].split('=')[-1].strip()])))

            elif 'Current cell volume' in line:
                snapshot['cell_volume'] = f90_float_parse(line.split('=')[1].split()[0].strip())
            elif 'Cell Contents' in line:
                snapshot['positions_frac'] = []
                snapshot['atom_types'] = []
                i = 1
                atoms = False
                while True:
                    if atoms:
                        if 'xxxxxxxxx' in flines[line_no + i]:
                            atoms = False
                            break
                        else:
                            temp_frac = flines[line_no + i].split()[3:6]
                            snapshot['positions_frac'].append(list(map(f90_float_parse, temp_frac)))
                            snapshot['atom_types'].append(flines[line_no + i].split()[1])
                    if 'x------' in flines[line_no + i]:
                        atoms = True
                    i += 1
                for ind, pos in enumerate(snapshot['positions_frac']):
                    for k in range(3):
                        if pos[k] > 1 or pos[k] < 0:
                            snapshot['positions_frac'][ind][k] %= 1

                snapshot['num_atoms'] = len(snapshot['positions_frac'])
                snapshot['stoichiometry'] = get_stoich(snapshot['atom_types'])

            # don't check if final_energy exists, as this will update for each GO step
            elif 'Final energy, E' in line or 'Final energy =' in line:
                snapshot['total_energy'] = f90_float_parse(line.split('=')[1].split()[0])
            elif 'Final free energy' in line:
                snapshot['smeared_free_energy'] = f90_float_parse(line.split('=')[1].split()[0])
            elif '0K energy' in line:
                snapshot['0K_energy'] = f90_float_parse(line.split('=')[1].split()[0])
            elif ' Forces **' in line:
                snapshot['forces'] = []
                i = 1
                max_force = 0
                forces = False
                while True:
                    if forces:
                        if '*' in flines[line_no + i].split()[1]:
                            forces = False
                            break
                        else:
                            force_on_atom = 0
                            snapshot['forces'].append([])
                            for j in range(3):
                                temp = flines[line_no + i].replace('(cons\'d)', '')
                                force_on_atom += f90_float_parse(temp.split()[3 + j])**2
                                snapshot['forces'][-1].append(f90_float_parse(temp.split()[3 + j]))
                            if force_on_atom > max_force:
                                max_force = force_on_atom
                    elif 'x' in flines[line_no + i]:
                        i += 1  # skip next blank line
                        forces = True
                    i += 1
                snapshot['max_force_on_atom'] = pow(max_force, 0.5)
            elif 'Stress Tensor' in line:
                i = 1
                while i < 20:
                    if 'Cartesian components' in flines[line_no + i]:
                        snapshot['stress'] = []
                        for j in range(3):
                            snapshot['stress'].append(list(map(f90_float_parse, (flines[line_no + i + j + 4].split()[2:5]))))
                    elif 'Pressure' in flines[line_no + i]:
                        snapshot['pressure'] = f90_float_parse(flines[line_no + i].split()[-2])
                        break
                    i += 1

            # use only finished iterations for counting number of complete GO steps
            elif ': finished iteration' in line and 'with enthalpy' in line:
                # don't include the "zeroth" step before anything has been moved
                if '0' not in line.split():
                    num_opt_steps += 1

        except Exception as exc:
            msg = 'Error on line {}, contents: {}, error: {}"'.format(line_no, line, exc)
            raise RuntimeError(msg)

    return intermediates, num_opt_steps


def _castep_scrape_beef(flines, castep):
    """ Scrape the Bayesian error estimate output from CASTEP, storing
    it under the `_beef` key.

    Parameters:
        flines (list): CASTEP output flines to scrape.
        castep (dict): dictionary to update with BEEF output.

    """
    from matador.utils.chem_utils import HARTREE_TO_EV
    for line_no, line in enumerate(flines):
        if 'Bayesian Error Estimate (BEE)' in line.strip():
            beef_start = line_no
            break
    else:
        return

    castep['_beef'] = {'thetas': [], 'xc_energy': [], 'total_energy': [],
                       'total_energy_per_atom': [], 'xc_energy_per_atom': []}

    for line_no, line in enumerate(flines[beef_start:]):
        if '<-- BEEF' in line:
            castep['_beef']['thetas'].append([f90_float_parse(val) for val in line.strip().split()[1:4]])
            castep['_beef']['xc_energy'].append(HARTREE_TO_EV*f90_float_parse(line.strip().split()[4]))
            castep['_beef']['total_energy'].append(HARTREE_TO_EV*f90_float_parse(line.strip().split()[6]))
            castep['_beef']['xc_energy_per_atom'].append(castep['_beef']['xc_energy'][-1] / castep['num_atoms'])
            castep['_beef']['total_energy_per_atom'].append(castep['_beef']['total_energy'][-1] / castep['num_atoms'])
        if 'BEEF completed' in line:
            break
    else:
        raise RuntimeError('End of BEEF estimate not found.')

    castep['_beef']['mean_total_energy'] = np.mean(castep['_beef']['total_energy'])
    castep['_beef']['mean_total_energy_per_atom'] = np.mean(castep['_beef']['total_energy']) / castep['num_atoms']
    castep['_beef']['std_dev_total_energy'] = np.std(castep['_beef']['total_energy'])
    castep['_beef']['std_dev_total_energy_per_atom'] = np.std(castep['_beef']['total_energy']) / castep['num_atoms']


def _castep_scrape_devel_code(flines, castep):
    """ Scrape the contents of the developer code block and
    extract any information about nanotube encapsulation.
    Searches for then scrapes the last occurence of "Developer
    Code". If nanotube information is found, the `encapsulated`
    flag is set to True.

    Parameters:
        flines (list): CASTEP output flines to scrape.
        castep (dict): dictionary to update with BEEF output.

    """
    last_devel_code = len(flines)
    for line_no, line in enumerate(flines):
        if line.startswith(' *') and '**** Developer Code ****' in line:
            last_devel_code = line_no

    castep['devel_code'] = []
    for line_no, line in enumerate(flines[last_devel_code+1:]):
        # look for end of block
        if '*************' in line.strip():
            break

        line = line.strip()
        if line:
            castep['devel_code'].append(line)

    devel_code = '\n'.join(castep['devel_code'])
    if 'gaussian_cylinder' in devel_code.lower():
        castep['encapsulated'] = True
        for line in castep['devel_code']:
            if 'radius' in line.lower():
                castep['cnt_radius'] = float(line.strip().split()[-1])

    # store as string rather than list
    castep['devel_code'] = devel_code


def _optados_get_projector_labels(header):
    """ Get OptaDOS projector labels from a pdos.dat or pdis.dat file,
    returning None-padded labels as `('species', 'ang_mom', 'spin channel')`,
    e.g. ('Li', 's', None) or ('Cr', None, 'up').

    Parameters:
        header (str): the file header containing projector labels.

    Returns:
        list(tuple(str, str, str)): projector labels formatted as above.

    """
    projectors = []
    for ind, line in enumerate(header):
        separators = ['Projector:', 'Column:']
        if any(sep in line for sep in separators):
            is_spin_pdos = 'Spin' in header[ind+1]
            # skip current line and column headings
            j = 2
            elements = []
            ang_mom_channels = []
            spin_channels = []
            # loop over file finding projector header blocks
            while ind + j + 1 < len(header) and not any(keyword in header[ind+j+1] for keyword in separators):
                elements.append(header[ind + j].split()[1])
                ang_mom_channels.append(header[ind + j].split()[3])
                if is_spin_pdos:
                    spin_channels.append(header[ind + j].split()[4].lower())
                j += 1

            projector_label = []

            # check that this projector contains exactly 1 species
            if len(set(elements)) == 1:
                projector_label.append(elements[0])
            else:
                projector_label.append(None)

            # check that this projector has exactly 1 ang mom channel
            if len(set(ang_mom_channels)) == 1:
                projector_label.append(ang_mom_channels[0])
            else:
                projector_label.append(None)

            # check that this projector has exactly 1 spin channel
            if len(set(spin_channels)) == 1:
                projector_label.append(spin_channels[0])
            else:
                projector_label.append(None)

            projector_label = tuple(projector_label)
            projectors.append(projector_label)

    return projectors


def get_seed_metadata(doc, seed):
    """ For a given document and seedname, look for
    ICSD CollCode, MaterialsProject IDs and DOIs to add
    to the document.

    Parameters:
        doc (dict): the input document.
        seed (str): the filename that is being scraped (sans file extension).

    """
    if '-CollCode-' in seed:
        doc['icsd'] = int(seed.split('CollCode-')[-1].split('-')[0].split('_')[0].split('.')[0])
    elif 'CollCode' in seed:
        doc['icsd'] = int(seed.split('CollCode')[-1].split('-')[0].split('_')[0].split('.')[0])
    elif '-ICSD-' in seed:
        doc['icsd'] = int(seed.split('-ICSD-')[-1].split('-')[0].split('_')[0].split('.')[0])
    if '-OQMD-' in seed:
        doc['oqmd_id'] = int(seed.split('-OQMD-')[-1].split('-')[0].split('_')[0].split('.')[0])
    elif '-OQMD_' in seed:
        doc['oqmd_id'] = int(seed.split('-OQMD_')[-1].split('-')[0].split('_')[0].split('.')[0])
    if '-MP-' in seed:
        doc['mp_id'] = int(seed.split('-MP-')[-1].split('-')[0].split('_')[0].split('.')[0])
    elif '-MP_' in seed:
        doc['mp_id'] = int(seed.split('-MP_')[-1].split('-')[0].split('_')[0].split('.')[0])
    if '-DOI-' in seed:
        doc['doi'] = seed.split('-DOI-')[-1].split('-')[0].replace('__', '/')



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functions to plot convergence test data
for a series of calculations.

"""

import glob
from decimal import Decimal, ROUND_UP
from collections import defaultdict
from traceback import print_exc
import numpy as np
from matador.plotting.plotting import plotting_function, SAVE_EXTS
from matador.scrapers.castep_scrapers import castep2dict
from matador.utils.chem_utils import get_formula_from_stoich

__all__ = ['plot_cutoff_kpt_grid']


@plotting_function
def plot_cutoff_kpt_grid(data, forces=False, max_energy=50, legend=True, colour_by='formula', **kwargs):
    """ Create a composite plot of cutoff energy and kpoint convergence.

    Parameters:
        data (dict): dictionary of convergence data.

    Keyword arguments:
        forces (bool): whether or not to plot forces.
        legend (bool): whether or not to show a legend.
        max_energy (float): max ylimit for convergence data in meV.

    """
    import matplotlib.pyplot as plt
    if forces:
        _, ax = plt.subplots(2, 2)
    else:
        _, ax = plt.subplots(1, 2)

    _, lines, labels = plot_field(
        data,
        field='formation_energy_per_atom',
        parameter='cut_off_energy',
        ax=ax[0],
        colour_by=colour_by,
        max_y=max_energy,
        reference='last'
    )

    if len(labels) < 30 and legend:
        plt.figlegend(lines, labels,
                      loc='upper center', fontsize=10, ncol=5)

    plot_field(
        data,
        field='formation_energy_per_atom',
        parameter='kpoints_mp_spacing',
        ax=ax[1],
        colour_by=colour_by,
        max_y=max_energy,
        reference='last'
    )

    if forces:
        plot_field(
            data,
            field='forces',
            parameter='cut_off_energy',
            ax=ax[2],
            reference='last'
        )
        plot_field(
            data,
            field='forces',
            parameter='kpoints_mp_spacing',
            ax=ax[3],
            reference='last'
        )

    if any(kwargs.get(ext) for ext in SAVE_EXTS):
        fname = 'conv'
        for ext in SAVE_EXTS:
            if kwargs.get(ext):
                plt.savefig('{}.{}'.format(fname, ext),
                            bbox_inches='tight', transparent=True)
                print('Wrote {}.{}'.format(fname, ext))

    plt.show()


def plot_field(data,
               field='formation_energy_per_atom', parameter='cut_off_energy',
               ax=None, reference='last', max_y=None, log=False, colour_by='formula'):
    """ Plot the convergence fields for each structure in `data` at
    each value of `parameter`.

    Parameters:
        data (dict): nested dictionary with keys for each structure
            that stores the convergence data to be plotted. e.g.
            {'structure_A':
                {'formation_energy_per_atom': [1, 2, 3],
                 'cutoff': [500, 400, 300]
                }
            }.

    Keyword arguments:
        field (str): the key under which convergence data is stored. This
            string is used for the y-axis label.
        parameter (str): the key under which the convergence parameter
            values are stored. This string is used for the x-axis label.
        ax (matplotlib.Axis): optional axis object to use for plot.
        reference (str): if 'last' then use the last value of the
            convergence data as the zero.

    """
    import matplotlib.pyplot as plt
    if ax is None:
        fig, ax = plt.subplots()

    lines = []
    labels = []
    labelled = []

    colourmap = {}
    if colour_by == 'formula':
        formulae = sorted(list({get_formula_from_stoich(data[key]['stoichiometry']) for key in data}))
        for formula in formulae:
            colourmap[formula] = next(ax._get_lines.prop_cycler)['color']

    for ind, key in enumerate(data):

        formula = get_formula_from_stoich(data[key]['stoichiometry'])
        try:
            values, parameters = get_convergence_values(data[key], parameter, field, reference=reference, log=log)
        except Exception:
            print('Missing data for {}->{}->{}'.format(key, parameter, field))
            continue

        label = None
        if colour_by == 'formula':
            c = colourmap[formula]
            if formula not in labelled:
                label = formula
                labelled.append(formula)
        else:
            c = next(ax._get_lines.prop_cycler)['color']
            label = key

        # only plot points if there's less than 25 lines
        if len(data) < 30:
            line, = ax.plot(parameters, values, 'o', markersize=5, alpha=1, label=label, lw=0, zorder=1000, c=c)
            points, = ax.plot(parameters, values, '-', alpha=0.2, lw=1, zorder=1000, c=c)
        else:
            points, = ax.plot(parameters, values, '-', alpha=0.5, lw=1, zorder=1000, label=label, c=c)
            line = points

        if label is not None:
            lines.append(line)
            labels.append(label)

    ax.set_xlabel(parameter.replace('_', ' '))
    ax.set_ylabel('Absolute relative {} (meV)'.format(field.replace('_', ' ')))
    if max_y is not None:
        ax.set_ylim(0, max_y)

    return ax, lines, labels


def round(n, prec):
    """ Replace default (bankers) rounding with "normal" rounding."""
    if prec is None:
        return n
    else:
        return float(Decimal(str(n)).quantize(Decimal('0.05'), rounding=ROUND_UP))


def get_convergence_files(path, only=None):
    """ Find all CASTEP files in the directory. """
    structure_files = defaultdict(list)
    files = glob.glob(path + '/*.castep')
    for file in files:
        if only is None or only in file:
            castep_dict, success = castep2dict(file, db=False)
            if not success:
                print('Failure to read castep file {}'.format(file))
            else:
                source = castep_dict['source'][0].split('/')[-1]
                source = source.replace('.castep', '')
                source = ''.join(source.split('_')[:-1])
                structure_files[source].append(castep_dict)

    return structure_files


def get_convergence_data(structure_files, conv_parameter='cut_off_energy', species=None):
    """ Parse cutoff energy/kpt spacing convergence calculations from list of files.

    Parameters:
        structure_files (list): list of filenames.

    Keyword arguments:
        conv_parameter (str): field for convergence parameter.
        species (list): only include structures with these species included.

    """
    scraped_from_filename = None

    form_set = {get_formula_from_stoich(structure_files[structure][0]['stoichiometry']) for structure in structure_files}
    if len(form_set) == 1:
        chempot_mode = False
        single = True
        print('Working in single stoichiometry mode..')

    else:
        print('Searching for chemical potentials')
        chempot_mode = True
        single = False
        chempots_dict = defaultdict(dict)

    if conv_parameter == 'kpoints_mp_spacing':
        rounding = None
    else:
        rounding = None

    data = {}

    if chempot_mode:
        for key in structure_files:
            skip = False
            if species is not None:
                for elem in structure_files[key][0]['stoichiometry']:
                    if elem[0] not in species:
                        skip = True
            if skip:
                continue
            for doc in structure_files[key]:
                scraped_from_filename = None
                if not single and len(doc['stoichiometry']) == 1:
                    if conv_parameter == 'kpoints_mp_spacing':
                        scraped_from_filename = float(doc['source'][0].split('/')[-1].split('_')[-1].split('A')[0])

                    if scraped_from_filename is not None:
                        rounded_field = round(scraped_from_filename, rounding)
                    else:
                        rounded_field = round(doc[conv_parameter], rounding)
                    energy_key = 'total_energy_per_atom'
                    chempots_dict[str(rounded_field)][doc['atom_types'][0]] = doc[energy_key]

    elems = set()

    if chempot_mode:
        for value in chempots_dict:
            for elem in chempots_dict[value]:
                elems.add(elem)

        for value in chempots_dict:
            for elem in elems:
                if elem not in chempots_dict[value]:
                    print('WARNING: {} chemical potential missing at {} = {} eV, skipping this value.'
                          .format(elem, conv_parameter, value))

    for key in structure_files:
        skip = False
        if species is not None:
            for elem in structure_files[key][0]['stoichiometry']:
                if elem[0] not in species:
                    skip = True
        if skip:
            continue

        data[key] = {}
        data[key][conv_parameter] = defaultdict(list)

        for doc in structure_files[key]:
            if conv_parameter == 'kpoints_mp_spacing':
                scraped_from_filename = float(doc['source'][0].split('/')[-1].split('_')[-1].split('A')[0])
            try:
                doc['formation_energy_per_atom'] = doc['total_energy_per_atom']
                if scraped_from_filename is not None:
                    rounded_field = round(scraped_from_filename, rounding)
                else:
                    rounded_field = round(doc[conv_parameter], rounding)
                if chempot_mode and len(doc['stoichiometry']) > 1:
                    for atom in doc['atom_types']:
                        doc['formation_energy_per_atom'] -= chempots_dict[str(rounded_field)][atom] / len(doc['atom_types'])

                data[key][conv_parameter][conv_parameter].append(rounded_field)
                data[key]['stoichiometry'] = doc['stoichiometry']
                data[key][conv_parameter]['formation_energy_per_atom'].append(doc['formation_energy_per_atom'])

                if 'forces' in doc:
                    data[key][conv_parameter]['forces'].append(np.sqrt(np.sum(np.asarray(doc['forces'])**2, axis=-1)))

            except Exception:
                print_exc()
                print('Error with {}'.format(key))

    if conv_parameter == 'kpoints_mp_spacing':
        reverse = True
    else:
        reverse = False

    for key in data:
        inds = [ind for ind, _ in sorted(enumerate(data[key][conv_parameter][conv_parameter]),
                                         key=lambda x: x[1], reverse=reverse)]
        for field in data[key][conv_parameter]:
            if field != 'stoichiometry':
                data[key][conv_parameter][field] = [data[key][conv_parameter][field][ind] for ind in inds]

    return data


def get_convergence_values(data, parameter, field, reference='last', log=False):
    """ Extract the data to plot for the given dictionary. """

    values = data[parameter][field]
    parameters = data[parameter][parameter]
    if len(values) != len(parameters):
        raise RuntimeError('Mismatched convergence data for {}->{}: {}'.format(field, parameter, data))

    values = np.asarray(values)
    parameters = np.asarray(parameters)

    if reference == 'last':
        values -= values[-1]
        values *= 1000
        values = np.abs(values)

    if log:
        with np.errstate(divide='ignore'):
            values = np.log10(values)

    return values, parameters


def combine_convergence_data(data_A, data_B):
    """ Combine dictionaries with potentially overlapping keys. """
    data = {}
    for key in data_A:
        data[key] = data_A[key]
        if key in data_B:
            data[key].update(data_B[key])
    for key in data_B:
        if key not in data:
            data[key] = data_B[key]

    return data



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the QuantumEspressoCalculator class for use
with the compute module.

"""

from matador.calculators import Calculator


class QuantumEspressoCalculator(Calculator):
    raise NotImplementedError



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The scrapers module implements the construction of Python dicts from
many common DFT/crystallography file formats.

Currently supported filetypes:

    - CASTEP filetypes: .cell, .param, .castep, .bands, .phonon, .usp
    - Custom .res type based on SHELX files, first used by AIRSS package https://www.mtg.msm.cam.ac.uk/Codes/AIRSS
    - OptaDOS output: .odo, .adaptive.dat
    - Quantum Espresso output files: .out
    - Magres files: .magres
    - Crystallographic Information File (via ASE): .cif

"""


__all__ = ['res2dict', 'cell2dict', 'param2dict', 'castep2dict', 'bands2dict', 'arbitrary2dict',
           'phonon2dict', 'phonon_dos2dict', 'optados2dict', 'usp2dict', 'pwout2dict', 'magres2dict', 'cif2dict']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'

from matador.scrapers.castep_scrapers import res2dict, cell2dict, param2dict
from matador.scrapers.castep_scrapers import castep2dict, bands2dict, arbitrary2dict
from matador.scrapers.castep_scrapers import phonon2dict, phonon_dos2dict, optados2dict, usp2dict
from matador.scrapers.qe_scrapers import pwout2dict
from matador.scrapers.magres_scrapers import magres2dict
from matador.scrapers.cif_scraper import cif2dict



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements a simple agnostic interface to various
queueing systems.

"""

import os
import abc


class QueueManager(abc.ABC):
    """ Abstract base class for queue managers. """
    token = None

    def __repr__(self):
        return '{} with env: {}'.format(type(self), self.env)

    def __init__(self):
        self.env = get_queue_env(self.token)

    @property
    def walltime(self):
        """ Return the allotted walltime in seconds,
        returning None if not available.

        """
        try:
            return self.get_walltime()
        except Exception:
            None

    @property
    def ntasks(self):
        return self.get_ntasks()

    @property
    def max_memory(self):
        """ Return the allotted memory in MB,
        returning None if not available.

        """
        try:
            return self.get_max_memory()
        except Exception:
            return None

    @property
    def array_id(self):
        return self.get_array_id()

    @abc.abstractmethod
    def get_walltime(self):
        pass

    @abc.abstractmethod
    def get_max_memory(self):
        pass

    @abc.abstractmethod
    def get_array_id(self):
        pass


def get_queue_env(token):
    """ Read os.environment variables for either PBS or SLURM
    prefixes, and return a dictionary of those vars only.

    Parameter:
        token (str): choose one of either SLURM or PBS explicitly.

    Returns:
        (dict, str): dictionary of keys from the detected/specified queue, and
            a string containing either "slurm" or "pbs".

    """
    return {key: os.environ[key] for key in os.environ if key.lower().startswith(token)}


def get_queue_manager():
    """ Detects whether PBS, SLURM or neither is being used
    by probing the environment variables SLURM_NTASKS and
    PBS_TASKNUM.

    Returns:
        str or None: either "slurm", "pbs" or None.

    Raises:
        SystemExit: if both SLURM and PBS were found.

    """
    queue_mgr = []
    if os.environ.get('SLURM_NTASKS') is not None:
        queue_mgr.append('slurm')
    if os.environ.get('PBS_TASKNUM') is not None:
        queue_mgr.append('pbs')

    if len(queue_mgr) > 1:
        raise RuntimeError('Both SLURM and PBS were found... aah!')
    if not queue_mgr:
        return None

    if queue_mgr[0] == 'slurm':
        from matador.compute.slurm import SlurmQueueManager
        return SlurmQueueManager()

    if queue_mgr[0] == 'pbs':
        from matador.compute.pbs import PBSQueueManager
        return PBSQueueManager()



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The compute module contains three submodules, compute, batch and
slurm.

The compute submodule contains the :class:`ComputeTask` class for performing
continually restarted geometry optimisation and SCF calculations in
CASTEP, as well as the execution of arbitrary programs with mpirun.

The batch submodule contains the :class:`BatchRun` class for running several
independent :class:`ComputeTask` instances on a folder of structures, without
clashes.

The slurm submodule provides a wrapper to useful slurm commands, and to
writing slurm job submission files.

"""


__all__ = ['ComputeTask', 'BatchRun', 'reset_job_folder']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.compute.compute import ComputeTask
from matador.compute.batch import BatchRun, reset_job_folder



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the base class Spatula that calls the scrapers
and interfaces with the MongoDB client.

"""


import os
import collections
import tempfile
import random
import datetime
import copy
import traceback as tb
import logging
import sys

import pymongo as pm

from matador.scrapers.castep_scrapers import castep2dict, param2dict, cell2dict
from matador.scrapers.castep_scrapers import res2dict
from matador.utils.cell_utils import calc_mp_spacing
from matador.utils.chem_utils import get_root_source
from matador.utils.cursor_utils import recursive_get
from matador.db import make_connection_to_collection
from matador.utils.db_utils import WORDS, NOUNS
from matador.config import load_custom_settings
from matador.utils.cursor_utils import loading_bar
from matador import __version__


class Spatula:
    """ The Spatula class implements methods to scrape folders and
    individual files for crystal structures and create a MongoDB
    document for each.

    Files types that can be read are:

        * CASTEP `.castep` output
        * SHELX (from airss.pl / pyAIRSS) `.res` output
        * CASTEP `.param`, `.cell` input

    This class will recursively scan directories from the cwd to
    find the files types above. Base filenames will be matched
    to prevent duplication of data from e.g. `.castep` and `.res`
    files. The following directory structures are recommended:

    - One `.res` file per structure and template `.cell` and
    `.param` files that provide all CASTEP parameters that structures
    in this folder were ran at.

    - One `.castep` file per structure, containing all information. If
    pseudopotential information is not present in the CASTEP file, this
    class will check for the corresponding `.usp` files and try to scrape
    those.

    """
    def __init__(self, *args, settings=None):
        """ Set up arguments and initialise DB client.

        Notes:
            Several arguments can be passed to this class from the command-line,
            and here are interpreted through *args:

        Parameters:
            db (str): the name of the collection to import to.
            scan (bool): whether or not to just scan the directory, rather
                than importing (automatically sets dryrun to true).
            dryrun (bool): perform whole process, up to actually importing to
                the database.
            tags (str): apply this tag to each structure added to database.
            force (bool): override rules about which folders can be imported
                into main database.
            recent_only (bool): if true, sort file lists by modification
                date and stop scanning when a file that already exists in
                database is found.

        """
        self.args = args[0]
        self.dryrun = self.args.get('dryrun')
        self.scan = self.args.get('scan')
        self.recent_only = self.args.get('recent_only')
        if self.scan:
            self.dryrun = True
        self.debug = self.args.get('debug')
        self.verbosity = self.args.get('verbosity') or 0
        self.config_fname = self.args.get('config')
        self.tags = self.args.get('tags')
        self.prototype = self.args.get('prototype')
        self.tag_dict = dict()
        self.tag_dict['tags'] = self.tags
        self.import_count = 0
        self.skipped = 0
        self.exclude_patterns = ['bad_castep', 'input']
        self.errors = 0
        self.struct_list = []
        self.path_list = []

        self.log = logging.getLogger('spatula')
        loglevel = {3: "DEBUG", 2: "INFO", 1: "WARN", 0: "ERROR"}
        self.log.setLevel(loglevel.get(self.verbosity, "WARN"))

        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        self.log.addHandler(handler)

        # I/O files
        if not self.dryrun:
            self.num_words = len(WORDS)
            self.num_nouns = len(NOUNS)

        if not self.scan:
            self.logfile = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
            self.manifest = tempfile.NamedTemporaryFile(mode='w+t', delete=False)

        if settings is None:
            self.settings = load_custom_settings(config_fname=self.config_fname,
                                                 debug=self.debug,
                                                 no_quickstart=self.args.get('no_quickstart'))
        else:
            self.settings = settings

        result = make_connection_to_collection(self.args.get('db'),
                                               check_collection=False,
                                               import_mode=True,
                                               override=self.args.get('no_quickstart'),
                                               mongo_settings=self.settings)

        self.client, self.db, self.collections = result
        # perform some relevant collection-dependent checks
        assert len(self.collections) == 1, 'Can only import to one collection.'
        self.repo = list(self.collections.values())[0]

        # if trying to import to the default repo, without doing a dryrun or forcing it, then
        # check if we're in the protected directory, i.e. the only one that is allowed to import
        # to the default collection
        default_collection = recursive_get(self.settings, ['mongo', 'default_collection'])
        try:
            default_collection_file_path = recursive_get(self.settings, ['mongo', 'default_collection_file_path'])
        except KeyError:
            default_collection_file_path = None

        if self.args.get('db') is None or self.args.get('db') == default_collection:
            if not self.dryrun and not self.args.get('force'):
                # if using default collection, check we are in the correct path
                if default_collection_file_path is not None:
                    if not os.getcwd().startswith(os.path.expanduser(default_collection_file_path)):
                        print(80 * '!')
                        print('You shouldn\'t be importing to the default database from this folder! '
                              'Please use --db <YourDBName> to create a new collection, '
                              'or copy these files to the correct place!')
                        print(80 * '!')
                        raise RuntimeError('Failed to import to default collection from '
                                           'current directory, import must be called from {}'
                                           .format(default_collection_file_path))
        else:
            if self.args.get('db') is not None:
                if any(['oqmd' in db for db in self.args.get('db')]):
                    exit('Cannot import directly to oqmd repo')
                elif len(self.args.get('db')) > 1:
                    exit('Can only import to one collection.')

        num_prototypes_in_db = self.repo.count_documents({'prototype': True})
        num_objects_in_db = self.repo.count_documents({})
        self.log.info("Found {} existing objects in database.".format(num_objects_in_db))
        if self.args.get('prototype'):
            if num_prototypes_in_db != num_objects_in_db:
                raise SystemExit('I will not import prototypes to a non-prototype database!')
        else:
            if num_prototypes_in_db != 0:
                raise SystemExit('I will not import DFT calculations into a prototype database!')

        if not self.dryrun:
            # either drop and recreate or create spatula report collection
            self.db.spatula.drop()
            self.report = self.db.spatula

        # scan directory on init
        self.file_lists = self._scan_dir()
        # if import, as opposed to rebuild, scan for duplicates and remove from list
        if not self.args.get('subcmd') == 'rebuild':
            self.file_lists, skipped = self._scan_dupes(self.file_lists)
            self.skipped += skipped

        # print number of files found
        self._display_import()

        # only create dicts if not just scanning
        if not self.scan:
            # convert to dict and db if required
            self._files2db(self.file_lists)
        if self.import_count == 0:
            print('No new structures imported!')
        if not self.dryrun and self.import_count > 0:
            print('Successfully imported', self.import_count, 'structures!')
            # index by enthalpy for faster/larger queries
            count = 0
            for _ in self.repo.list_indexes():
                count += 1
            # ignore default id index
            if count > 1:
                self.log.info('Index found, rebuilding...')
                self.repo.reindex()
            else:
                self.log.info('Building index...')
                self.repo.create_index([('enthalpy_per_atom', pm.ASCENDING)])
                self.repo.create_index([('stoichiometry', pm.ASCENDING)])
                self.repo.create_index([('cut_off_energy', pm.ASCENDING)])
                self.repo.create_index([('species_pot', pm.ASCENDING)])
                self.repo.create_index([('kpoints_mp_spacing', pm.ASCENDING)])
                self.repo.create_index([('xc_functional', pm.ASCENDING)])
                self.repo.create_index([('elems', pm.ASCENDING)])
                # index by source for rebuilds
                self.repo.create_index([('source', pm.ASCENDING)])
                self.log.info('Done!')
        elif self.dryrun:
            self.log.info('Dryrun complete!')

        if not self.scan:
            self.logfile.seek(0)
            errors = sum(1 for line in self.logfile)
            self.errors += errors
            if errors == 1:
                self.log.warning('There is 1 error to view in {}'.format(self.logfile.name))
            elif errors == 0:
                self.log.warning('There are no errors to view in {}'.format(self.logfile.name))
            elif errors > 1:
                self.log.warning('There are {} errors to view in {}'.format(errors, self.logfile.name))

        try:
            self.logfile.close()
        except Exception:
            pass
        try:
            self.manifest.close()
        except Exception:
            pass

        if not self.dryrun:
            # construct dictionary in spatula_report collection to hold info
            report_dict = dict()
            report_dict['last_modified'] = datetime.datetime.utcnow().replace(microsecond=0)
            report_dict['num_success'] = self.import_count
            report_dict['num_errors'] = errors
            report_dict['version'] = __version__
            self.report.insert_one(report_dict)

    def _struct2db(self, struct):
        """ Insert completed Python dictionary into chosen
        database, with generated text_id. Add quality factor
        for any missing data.

        Parameters:
            struct (dict): dictionary containing structure.

        Returns:
            int: 1 if successfully inputted into database, 0 otherwise.

        """
        try:
            plain_text_id = [WORDS[random.randint(0, self.num_words-1)].strip(),
                             NOUNS[random.randint(0, self.num_nouns-1)].strip()]
            struct['text_id'] = plain_text_id
            if 'tags' in self.tag_dict:
                struct['tags'] = self.tag_dict['tags']
            struct['quality'] = 5
            # if any missing info at all, score = 0
            # include elem set for faster querying
            if 'elems' not in struct:
                struct['elems'] = sorted(list(set(struct['atom_types'])))

            # check basic DFT params if we're not in a prototype DB
            if not self.args.get('prototype'):
                failed_checks = []
                if 'species_pot' not in struct:
                    struct['quality'] = 0
                    failed_checks.append('missing all pspots')
                else:
                    specified = []
                    for elem in struct['stoichiometry']:
                        # remove all points for a missing pseudo
                        if elem[0] not in struct['species_pot']:
                            struct['quality'] = 0
                            failed_checks.append('missing pspot for {}'.format(elem[0]))
                        else:
                            specified.append(elem[0])
                            # remove a point for a generic OTF pspot
                            if 'OTF' in struct['species_pot'][elem[0]].upper():
                                struct['quality'] -= 1
                                failed_checks.append('pspot not fully specified for {}'.format(elem[0]))
                    struct['species_pot'] = {species: struct['species_pot'][species] for species in specified}

                if 'xc_functional' not in struct:
                    struct['quality'] = 0
                    failed_checks.append('missing xc functional')

                if 'cut_off_energy' not in struct:
                    struct['quality'] -= 1
                if 'kpoints_mp_spacing' not in struct:
                    struct['quality'] -= 1

            else:
                struct['prototype'] = True
                struct['xc_functional'] = 'xxx'
                struct['enthalpy_per_atom'] = 0
                struct['enthalpy'] = 0
                struct['pressure'] = 0
                struct['species_pot'] = {}

            root_src = get_root_source(struct)
            struct['root_source'] = root_src
            exts = ['.castep', '.res', '.history', '.history.gz']
            for ext in exts:
                for src in struct['source']:
                    if src.endswith(ext):
                        expanded_root_src = src

            if struct['quality'] == 5:
                struct_id = self.repo.insert_one(struct).inserted_id
                self.struct_list.append((struct_id, root_src))
                self.manifest.write('+ {}\n'.format(expanded_root_src))
                if self.debug:
                    print('Inserted', struct_id)
            else:
                self.logfile.write('? {} failed quality checks: {}\n'.format(expanded_root_src, failed_checks))
                if self.debug:
                    print('Error with', root_src)
                return 0

        except Exception as exc:
            # this shouldn't fail, but if it does, fail loudly but cleanly
            self.logfile.write('! Importer produced an unexpected error {}. Final state of struct: {}\n'.format(exc, struct))
            tb.print_exc()
            return 0

        return 1

    def _files2db(self, file_lists):
        """ Take all files found by scan and appropriately create dicts
        holding all available data; optionally push to database. There are
        three possible routes this function will take:

            - AIRSS-style results, e.g. folder of res files, cell and param,
                will be scraped into a single dictionary per structure.
            - Straight CASTEP results will be scraped into one structure per
                .castep file.
            - Prototype-style results will take res files and turn them
                into simple structural data entries, without DFT parameters.

        Parameters:
            file_lists (dict): filenames and filetype counts stored by directory name key.

        """
        print('\n{:^52}'.format('###### RUNNING IMPORTER ######') + '\n')
        for _, root in enumerate(file_lists):
            root_str = root
            if root_str == '.':
                root_str = os.getcwd().split('/')[-1]
            self.log.info("Dictifying directory: {}".format(root_str))

            if self.args.get('prototype'):
                self.import_count += self._scrape_prototypes(file_lists, root)

            else:
                # default to only scraping castep files
                style = 'castep'
                # if there are multiple res files, per cell, assume we are working in "airss" mode
                if file_lists[root]['res_count'] > 0:
                    if (file_lists[root]['castep_count'] < file_lists[root]['res_count'] and
                            file_lists[root]['cell_count'] <= file_lists[root]['res_count']):
                        style = 'airss'
                if style == 'airss':
                    imported = self._scrape_multi_file_results(file_lists, root)
                # otherwise, we are just in a folder of CASTEP files
                elif style == 'castep':
                    imported = self._scrape_single_file_structures(file_lists, root)

            self.import_count += imported
            if imported > 0:
                print('Imported {} structures from {}'.format(imported, root))
                self.path_list.append(root)

        if self.struct_list and not self.dryrun:
            self._update_changelog()

    def _scrape_multi_file_results(self, file_lists, root):
        """ Add structures to database by parsing .res or .castep files., with DFT data
        scraped from .castep/.cell/.param files in the same folder, i.e. data from multiple
        files.

        Parameters:
            file_lists: dictionary containing counts of file types in each
                sub-directory.
            root: name of sub-directory.

        Returns:
            int: number of structures successfully imported.

        """
        multi = False  # are there multiple param/cell files?
        cell = False  # was the cell file successfully scraped?
        param = False  # was the param file successfully scraped?
        import_count = 0  # how many files have been successfully imported?
        success = False

        if file_lists[root]['param_count'] == 1:
            param_dict, success = param2dict(file_lists[root]['param'][0],
                                             debug=self.debug, noglob=True,
                                             verbosity=self.verbosity)
            param = success
            if not success:
                self.logfile.write(param_dict)
        elif file_lists[root]['param_count'] > 1:
            self.log.warning('Multiple param files found: {}'.format(file_lists[root]['param']))
            multi = True
        if file_lists[root]['cell_count'] == 1:
            cell_dict, success = cell2dict(file_lists[root]['cell'][0],
                                           db=True,
                                           debug=self.debug,
                                           noglob=True,
                                           verbosity=self.verbosity)
            cell = success
            if not success:
                self.logfile.write(str(cell_dict))
        elif file_lists[root]['cell_count'] > 1:
            multi = True
            self.log.warning('Multiple param files found: {}'.format(file_lists[root]['cell']))
        if multi:
            found_multi = False
            for param_name in file_lists[root]['param']:
                for cell_name in file_lists[root]['cell']:
                    if param_name.split('.')[0] in cell_name:
                        cell_dict, success = cell2dict(cell_name,
                                                       debug=self.debug,
                                                       db=True,
                                                       verbosity=self.verbosity)
                        cell = success
                        if not success:
                            self.logfile.write(str(cell_dict))
                            continue

                        param_dict, success = param2dict(param_name,
                                                         debug=self.debug,
                                                         verbosity=self.verbosity)
                        param = success
                        if not success:
                            self.logfile.write(param_dict)

                        if success:
                            found_multi = True
                            self.log.info('Found matching cell and param files: {}'.format(param_name))
                            break

            if not found_multi:
                self.log.warning("Unable to find matching cell and param files for {}".format(root))

        # combine cell and param dicts for folder
        input_dict = dict()
        if cell and param:
            input_dict.update(cell_dict)
            input_dict.update(param_dict)
            input_dict['source'] = cell_dict['source'] + param_dict['source']
        else:
            self.logfile.write('! {} failed to scrape any cell and param \n'.format(root))

        # create res dicts and combine them with input_dict
        for _, _file in enumerate(loading_bar(file_lists[root]['res'], verbosity=self.verbosity)):
            exts_with_precedence = ['.castep', '.history', 'history.gz']
            # check if a castep-like file exists instead of scraping res
            if any([_file.replace('.res', ext) in file_lists[root]['castep']
                    for ext in exts_with_precedence]):
                for ext in exts_with_precedence:
                    if _file.replace('.res', ext) in file_lists[root]['castep']:
                        struct_dict, success = castep2dict(_file.replace('.res', ext),
                                                           debug=False, noglob=True,
                                                           dryrun=self.args.get('dryrun'),
                                                           verbosity=self.verbosity)
                        break
            # otherwise, scrape res file
            else:
                struct_dict, success = res2dict(_file, verbosity=self.verbosity, noglob=True)

            if not success:
                self.logfile.write('! {}'.format(struct_dict))
            else:
                try:
                    final_struct = copy.deepcopy(input_dict)
                    final_struct.update(struct_dict)
                    # calculate kpoint spacing if not found
                    if 'lattice_cart' not in final_struct and 'lattice_abc' not in final_struct:
                        msg = '! {} missing lattice'.format(_file)
                        self.logfile.write(msg)

                    if 'kpoints_mp_spacing' not in final_struct and 'kpoints_mp_grid' in final_struct:
                        final_struct['kpoints_mp_spacing'] = calc_mp_spacing(final_struct['lattice_cart'],
                                                                             final_struct['mp_grid'])
                    final_struct['source'] = struct_dict['source']
                    if 'source' in input_dict:
                        final_struct['source'] += input_dict['source']

                    if not self.dryrun:
                        final_struct.update(self.tag_dict)
                        import_count += self._struct2db(final_struct)

                except Exception as exc:
                    self.log.error('Unexpected error for {}, {}'.format(_file, final_struct))
                    raise exc

        return import_count

    def _scrape_single_file_structures(self, file_lists, root):
        """ Scrape data from one of CASTEP-like, .synth and .expt files,
        and push to database.

        Parameters:
            file_lists (dict): filenames and filetype counts stored by directory name key.
            root (str): directory name to scrape.

        Returns:
            int: number of successfully structures imported to database.

        """
        import_count = 0
        for _, file in enumerate(loading_bar(file_lists[root]['castep'], verbosity=self.verbosity)):
            castep_dict, success = castep2dict(file, debug=False, verbosity=self.verbosity)
            if not success:
                self.logfile.write('! {}'.format(castep_dict))
            else:
                final_struct = castep_dict
                if not self.dryrun:
                    final_struct.update(self.tag_dict)
                    import_count += self._struct2db(final_struct)

        return import_count

    def _scrape_prototypes(self, file_lists, root):
        """ Scrape prototype data, i.e. structures with no DFT data, and push
        to database.

        Parameters:
            file_lists (dict): filenames and filetype counts stored by directory name key.
            root (str): directory name to scrape.

        Returns:
            int: number of successfully structures imported to database.

        """
        import_count = 0
        for _, file in enumerate(file_lists[root]['res']):
            res_dict, success = res2dict(file, db=False, verbosity=self.verbosity, noglob=True)
            if not success:
                self.logfile.write('! {}'.format(res_dict))
            else:
                final_struct = res_dict
                if not self.dryrun:
                    final_struct.update(self.tag_dict)
                    import_count += self._struct2db(final_struct)

        return import_count

    def _update_changelog(self):
        """ Add a list of ObjectIds to a collection called __changelog_{collection_name},
        storing "commits" that can be undone or reverted to.

        """
        changes = {'date': datetime.datetime.today(),
                   'count': len(self.struct_list),
                   'id_list': [struct[0] for struct in self.struct_list],
                   'src_list': [struct[1] for struct in self.struct_list],
                   'path_list': self.path_list}
        self.db['__changelog_{}'.format(self.repo.name)].insert_one(changes)

    def _scan_dir(self):
        """ Scans folder topdir recursively, returning list of
        CASTEP/AIRSS input/output files.

        Returns:
            dict: dictionary keyed by directory name that stores filename
                and filetype counts.

        """
        self.log.info("Scanning for directories...")
        file_lists = dict()
        topdir = '.'
        topdir_string = os.getcwd().split('/')[-1]
        self.log.info('Scanning {} for CASTEP/AIRSS output files... '.format(topdir_string))
        for root, _, files in os.walk(topdir, followlinks=True, topdown=True):
            for pattern in self.exclude_patterns:
                if pattern in root:
                    self.log.debug('Skipping directory {} as it matched exclude pattern {}'.format(root, pattern))
            # get absolute path for rebuilds
            root = os.path.abspath(root)
            file_lists[root] = collections.defaultdict(list)
            file_lists[root]['res_count'] = 0
            file_lists[root]['cell_count'] = 0
            file_lists[root]['param_count'] = 0
            file_lists[root]['castep_count'] = 0
            file_lists[root]['synth_count'] = 0
            file_lists[root]['expt_count'] = 0
            for _file in files:
                _file = root + '/' + _file
                self.log.debug(_file)
                if _file.endswith('.res'):
                    file_lists[root]['res'].append(_file)
                    file_lists[root]['res_count'] += 1
                elif (_file.endswith('.castep') or _file.endswith('.history') or
                      _file.endswith('.history.gz')):
                    file_lists[root]['castep'].append(_file)
                    file_lists[root]['castep_count'] += 1
                elif _file.endswith('.cell'):
                    if _file.endswith('-out.cell'):
                        continue
                    else:
                        file_lists[root]['cell'].append(_file)
                        file_lists[root]['cell_count'] += 1
                elif _file.endswith('.param'):
                    file_lists[root]['param'].append(_file)
                    file_lists[root]['param_count'] += 1
                elif _file.endswith('.synth'):
                    file_lists[root]['synth'].append(_file)
                    file_lists[root]['synth_count'] += 1
                elif _file.endswith('.expt'):
                    file_lists[root]['expt'].append(_file)
                    file_lists[root]['expt_count'] += 1

        # finally, filter the castep list so that history/castep/history.gz duplicates
        # are not included
        for root in file_lists:
            cached_list = [entry.split('/')[-1].split('.')[0] for entry in file_lists[root]['castep']]
            duplicates = []
            for ind, file in enumerate(file_lists[root]['castep']):
                if not file.endswith('.castep') and cached_list.count(file.split('/')[-1].split('.')[0]) > 1:
                    duplicates.append(ind)
                    file_lists[root]['castep_count'] -= 1
            file_lists[root]['castep'] = [entry for ind, entry in
                                          enumerate(file_lists[root]['castep']) if ind not in duplicates]

        self.log.info("Scanning completed.")
        return file_lists

    def _scan_dupes(self, file_lists):
        """ Scan the file_lists made by scan_dir and remove
        structures already in the database by matching sources.

        Parameters:
            file_lists (dict): dict with directory name keys containing
                file names and filetype counts.

        Returns:
            dict: the input dict minus duplicates.

        """
        self.log.info("Scanning for duplicates...")
        skipped = 0
        new_file_lists = copy.deepcopy(file_lists)
        for _, root in enumerate(file_lists):
            # per folder delete list
            # do not import castep or res if seed name in db already
            delete_list = {}
            delete_list['castep'] = set()
            delete_list['res'] = set()
            types = ['castep', 'res']
            for structure_type in types:
                if self.recent_only:
                    try:
                        new_file_lists[root][structure_type].sort(key=lambda x: os.stat(x).st_ctime)
                    except Exception:
                        self.log.warning('Unable to sort files by creation date, ignoring `recent_only=True`')

                for file_ind, _file in enumerate(new_file_lists[root][structure_type]):
                    # find number of entries with same root filename in database
                    structure_exts = ['.castep', '.res', '.history', '.history.gz']
                    ext = [_ext for _ext in structure_exts if _file.endswith(_ext)]
                    assert len(ext) == 1
                    ext = ext[0]
                    structure_count = 0
                    for other_ext in structure_exts:
                        structure_count += self.repo.count_documents(
                            {'source': {'$in': [_file.replace(ext, other_ext)]}})

                    if structure_count > 1:
                        self.log.error('Found double in database, this needs to be dealt with manually: {}'.format(_file))

                    # if duplicate found, don't reimport
                    if structure_count >= 1:
                        # need to loop over possible file types to get correct fname
                        _add_to_delete_lists(_file, root, new_file_lists, delete_list)
                        if self.recent_only:
                            skipping = len(new_file_lists[root][structure_type][file_ind:])
                            skipped += skipping
                            for other_file in new_file_lists[root][structure_type][file_ind:]:
                                _add_to_delete_lists(other_file, root, new_file_lists, delete_list)
                            break

                        assert len(new_file_lists[root][structure_type]) == len(file_lists[root][structure_type]) + skipped

            for _type in types:
                new_file_lists[root][_type] = [_file for _file in new_file_lists[root][_type]
                                               if _file not in delete_list[_type]]

            for file_type in types:
                if len(new_file_lists[root][file_type]) != new_file_lists[root][file_type + '_count']:
                    raise RuntimeError('{} count {} does not match number of files {}'
                                       .format(file_type,
                                               len(new_file_lists[root][file_type]),
                                               new_file_lists[root][file_type + '_count']))

        self.log.info("Removed duplicates from import list...")
        return new_file_lists, skipped

    def _display_import(self):
        """ Display number of files to be imported and a breakdown of
        their types.

        """
        exts = ['res', 'cell', 'castep', 'param', 'synth', 'expt']
        counts = {ext: 0 for ext in exts}
        for root in self.file_lists:
            for ext in exts:
                counts[ext] += self.file_lists[root]['{}_count'.format(ext)]

        print('\n\n')
        if self.recent_only:
            print('\t\tSkipped {} files filtered by creation date (st_ctime).'.format(self.skipped))

        for ext in exts:
            print('\t\t{:8d}\t\t.{} files'.format(counts[ext], ext))


def _add_to_delete_lists(_file, root, new_file_lists, delete_list):
    """ Add files to the delete list, with the correct file types. """
    structure_exts = {'castep': 'castep', 'res': 'res', 'history': 'castep', 'history.gz': 'castep'}
    for _type in structure_exts:
        fname_trial = _file.replace(_file.split('.')[-1], _type)
        list_type = structure_exts[_type]
        if fname_trial in new_file_lists[root][list_type] and fname_trial not in delete_list[list_type]:
            delete_list[list_type].add(fname_trial)
            new_file_lists[root][list_type + '_count'] -= 1



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The workflows.castep submodule contains any workflows that rely on
CASTEP-specific implementation, for example electronic spectropscopy or phonons.

"""


__all__ = ['castep_full_phonon', 'castep_full_spectral', 'castep_elastic']

__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.workflows.castep.phonons import castep_full_phonon
from matador.workflows.castep.spectral import castep_full_spectral
from matador.workflows.castep.elastic import castep_elastic



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file defines some useful scraper functionality,
like custom errors and a scraper function wrapper.
"""


import glob
import traceback as tb

from matador.orm.spectral import VibrationalDOS, VibrationalDispersion
from matador.orm.spectral import ElectronicDOS, ElectronicDispersion
from matador.crystal import Crystal

MODEL_REGISTRY = {
    "phonon_dos2dict": VibrationalDOS,
    "phonon2dict": VibrationalDispersion,
    "optados2dict": ElectronicDOS,
    "bands2dict": ElectronicDispersion,
    "castep2dict": Crystal,
    "res2dict": Crystal,
    "cif2dict": Crystal,
}


def scraper_function(function):
    """ Wrapper for scraper functions to handle exceptions and
    template the scraper functions to work for multiples files
    at once.

    """

    from functools import wraps

    @wraps(function)
    def wrapped_scraper_function(*args, verbosity=1, fail_fast=False, **kwargs):
        """ Wrap and return the scraper function, handling the
        multiplicity of file names.

        """

        if kwargs.get('no_wrap'):
            return function(*args, **kwargs)

        result = None
        seed = args[0]
        if isinstance(seed, str):
            if '*' in seed and not kwargs.get('noglob'):
                seed = sorted(glob.glob(seed))
            else:
                seed = [seed]

        failures = []
        cursor = []

        if not seed:
            print('Nothing to scrape.')
            return

        for _seed in seed:
            # we can get away with this as each
            # scraper function only has one arg
            try:
                result, success = function(_seed, verbosity=verbosity, **kwargs)
            # UnicodeDecodeErrors require 5 arguments, so handle these separately
            except (FileNotFoundError, UnicodeError) as oops:
                raise oops
            except Exception as oops:
                success = False
                result = type(oops)('{}: {}\n'.format(_seed, oops))

                if verbosity >= 1:
                    msg = '{}: {} {}'.format(_seed, type(oops), oops)
                    print(msg)
                if verbosity >= 2:
                    tb.print_exc()

                if fail_fast:
                    raise oops

            if len(seed) == 1:
                if success and kwargs.get('as_model'):
                    orm = _as_model(result, function)
                    if orm is not None:
                        result = orm
                if not success and verbosity >= 1:
                    print("Failed to scrape file {}".format(seed))

                return result, success

            if not success:
                failures += [_seed]
            else:
                if kwargs.get('as_model'):
                    orm = _as_model(result, function, debug=kwargs.get('debug'))
                    cursor.append(orm)
                if not kwargs.get('as_model') or orm is None:
                    cursor.append(result)

        if verbosity >= 1:
            print("Successfully scraped {} out of {} files.".format(len(cursor), len(cursor) + len(failures)))

        return cursor, failures

    return wrapped_scraper_function


def _as_model(doc, function, debug=False):
    """ Convert the document to the appropriate orm model. """
    model = MODEL_REGISTRY.get(function.__name__)
    orm = None
    if model is not None:
        try:
            orm = model(doc)
        except Exception:
            print('Unable to convert scraped dict to model {}'.format(model.__name__))
    else:
        print('as_model keyword not supported for {}, not converting'.format(function.__name__))

    return orm


def f90_float_parse(val):
    """ Wrapper to float that handles Fortran's horrible behaviour for
    float exponents <= 100, e.g. 1e-100 -> 1.0000000-100 rather than
    1.000000E-100. Also handles "+" signs in front of numbers.

    Parameters:
        val (str): the string to cast to a float.

    """
    try:
        return float(val)
    except ValueError as exc:
        # if the E is being displayed, then something else has gone wrong
        if 'E' in val:
            raise exc
        # if there's a minus sign after the first char, but no E...
        if len(val) > 1 and '-' in val[1:]:
            val = val[0] + val[1:].replace('-', 'E-')
        if val.startswith('+'):
            val = val[1:]
        return float(val)


class DFTError(Exception):
    """ Quick DFT exception class for unconverged or
    non-useful calculations.

    """


class ComputationError(Exception):
    """ Raised when the calculation fails to do the DFT.
    Distinct from DFTError as this is an issue of numerics
    or chemistry, where this is raised for technical issues,
    e.g. CASTEP crashes.

    """
